from pathlib import Path

import pytest

from mckit.cli.runner import mckit
from mckit.parser import from_file
from mckit.utils.resource import filename_resolver

data_filename_resolver = filename_resolver("tests.cli")


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["transform"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_transformation_is_not_defined(runner):
    result = runner.invoke(
        mckit,
        args=["transform", data_filename_resolver("data/simple_cubes.mcnp")],
        catch_exceptions=False,
    )
    assert result.exit_code > 0
    assert "Missing option" in result.output


def test_output_is_not_defined(runner):
    result = runner.invoke(
        mckit,
        args=[
            "transform",
            "--transformation",
            "1",
            data_filename_resolver("data/simple_cubes.mcnp"),
        ],
        catch_exceptions=False,
    )
    assert result.exit_code > 0
    assert "Missing option" in result.output
    assert "output" in result.output


def test_not_existing_mcnp_file(runner):
    result = runner.invoke(
        mckit,
        args=["transform", "-t", "1", "not-existing.imcnp"],
        catch_exceptions=False,
    )
    assert result.exit_code > 0
    assert "Path 'not-existing.imcnp' does not exist" in result.output


@pytest.mark.parametrize(
    "msg, _source, transformation,, expected",
    [
        (
            "identical transformation",
            "data/simple_cubes_with_tallies.mcnp",
            "1",
            "data/simple_cubes_with_tallies.mcnp",
        ),
    ],
)
def test_happy_path(runner, msg, _source, transformation, expected):
    source = data_filename_resolver(_source)
    out = Path(source).name
    transformations = data_filename_resolver("data/brick-transformations.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            mckit,
            args=[
                "transform",
                "-t",
                transformation,
                "-i",
                transformations,
                "-o",
                str(out),
                source,
            ],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, "Should success"
        out_path = Path(out)
        assert out_path.exists()


@pytest.mark.parametrize(
    "source, transformation, transformations, expected",
    [
        (
            "data/brick1.mcnp",
            "1",
            "data/brick-transformations.txt",
            "data/brick1-x+10.mcnp",
        ),
        (
            "data/brick1.mcnp",
            "2",
            "data/brick-transformations.txt",
            "data/brick1-rot-z90.mcnp",
        ),
    ],
)
def test_when_transform_happy_path(runner, source, transformation, transformations, expected):
    source = data_filename_resolver(source)
    transformations = data_filename_resolver(transformations)
    with runner.isolated_filesystem():
        out = "test_when_transformation_is_specified_by_spec.out"
        result = runner.invoke(
            mckit,
            args=[
                "transform",
                "-o",
                out,
                "-t",
                transformation,
                "-i",
                transformations,
                source,
            ],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, f"Failed to transform {source}"
        expected_universe = from_file(data_filename_resolver(expected)).universe
        expected_surfaces = list(expected_universe[0])
        actual_universe = from_file(out).universe
        actual_surfaces = list(actual_universe[0])

        def key(a):
            return a.args[0].options["name"]

        assert sorted(actual_surfaces, key=key) == sorted(expected_surfaces, key=key)


# @pytest.mark.parametrize(
#     "source, expected",
#     [
#         (
#             "cli/data/simple_cubes.mcnp",
#             "title.txt \
#         cells_start.txt cells.txt cells_end.txt \
#         surfaces_start.txt surfaces.txt surfaces_end.txt \
#         materials_start.txt materials.txt materials_end.txt \
#         transformations_start.txt transformations_end.txt \
#         cards.txt \
#         new_line.txt",
#         )
#     ],
# )
# def test_when_separator_files_are_required(runner, source, expected):
#     source = data_filename_resolver(source)
#     with runner.isolated_filesystem():
#         result = runner.invoke(
#             mckit, args=["split", "--separators", source], catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         out = get_default_output_directory(source, ".split")
#         assert out.is_dir()
#         expected = expected.split()
#         for e in expected:
#             assert (out / e).exists()
#         text = (out / "cells_start.txt").read_text(encoding=MCNP_ENCODING)
#         assert is_comment_text(text), "Should be MCNP comment text"


# @pytest.mark.parametrize("source,expected", [
#     ("cli/data/simple_cubes.mcnp", "envelopes.i u1.i u2.i".split()),
# ])
# def test_when_only_source_is_specified(runner, source, expected):
#     source: Path = data_filename_resolver(source)
#     with runner.isolated_filesystem():
#         result = runner.invoke(mckit, args=['split', source], catch_exceptions=False)
#         assert result.exit_code == 0, \
#             "Should success without specified output: " + result.output
#         output: Path = get_default_output_directory(source)
#         for f in expected:
#             p = output / f
#             assert p.exists(), \
#                 f"Should store the file {p} in the default directory '{output}'"
#             model = mk.read_mcnp(p)
#             for cell in model:
#                 assert 'U' not in cell.options or cell.options['U'].name() == 0
#
#
# @pytest.mark.parametrize("source,output,expected", [
#     ("cli/data/simple_cubes.mcnp", "split-1", "envelopes.i u1.i u2.i".split()),
# ])
# def test_when_output_is_specified(runner, source, output, expected):
#     source = data_filename_resolver(source)
#     with runner.isolated_filesystem():
#         output = Path(output)
#         assert not output.exists(), f"The {output} directory should not exist before the test run"
#         result = runner.invoke(mckit, args=['split', "--output", str(output), source], catch_exceptions=False)
#         assert result.exit_code == 0
#         assert output.exists(), f"The {output} directory should exist after the test run"
#         for f in expected:
#             assert (output / f).exists(), \
#                 f"Should store the file {f} in the default directory 'universes'"
#
#
# def test_when_output_file_exists_and_override_is_not_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "universes/envelopes.i"
#         output.parent.mkdir(parents=True)
#         output.touch(exist_ok=False)
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code != 0, \
#             "Should fail when output file exists and --override is not specified"
#
#
# def test_when_output_file_exists_and_override_is_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "simple_cubes./envelopes.i"
#         output.parent.mkdir(parents=True)
#         output.touch(exist_ok=False)
#         result = runner.invoke(
#             mckit,
#             args=["--override", "split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, \
#             "Should success when output file exists and --override is specified"
#
#
#
# def test_fill_descriptor(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "universes/fill-descriptor.toml"
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         assert output.exists()
#         with output.open() as fid:
#             fill_descriptor = fid.read()
#             assert fill_descriptor.find('simple_cubes.mcnp')
#             fill_descriptor = tk.parse(fill_descriptor)
#             assert 'created' in fill_descriptor
#             assert '2' in fill_descriptor
#             assert 'universe' in fill_descriptor['2']
#             assert 1 == fill_descriptor['2']['universe']
#             assert 'u1.i' == fill_descriptor['2']['file']
#
#
#
# def test_fill_descriptor_when_fill_descriptor_file_is_specified(runner):
#     source = data_filename_resolver("cli/data/simple_cubes.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         fill_descriptor_path = Path(prefix) / "fill-descriptor-special.toml"
#         result = runner.invoke(
#             mckit,
#             args=["split", "--fill-descriptor", str(fill_descriptor_path), source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         assert fill_descriptor_path.exists()
#
#
# def test_anonymous_transformation(runner):
#     source = data_filename_resolver("cli/data/cubes_with_fill_transforms.mcnp")
#     with runner.isolated_filesystem() as prefix:
#         output = Path(prefix) / "cubes_with_fill_transforms.universes"
#         result = runner.invoke(
#             mckit,
#             args=["split", source],
#             catch_exceptions=False
#         )
#         assert result.exit_code == 0, "Should success"
#         with open(output / "fill-descriptor.toml") as fid:
#             descriptor = tk.parse(fid.read())
#             spec = descriptor['2']['transform']
#             assert float(spec[1]) == -1.0, "Fill descriptor is wrong"



from pathlib import Path

import pytest

from mckit.utils.resource import filename_resolver, path_resolver

THIS_FILENAME = Path(__file__).name


# noinspection PyCompatibility
@pytest.mark.parametrize(
    "package, resource, expected",
    [
        ("tests", "cli/data/simple_cubes.mcnp", "/cli/data/simple_cubes.mcnp"),
    ],
)
def test_filename_resolver(package, resource, expected):
    resolver = filename_resolver(package)
    actual = resolver(resource)
    assert actual.replace("\\", "/").endswith(expected), "Failed to compute resource file name"
    assert Path(actual).exists(), f"The resource {resource!r} is not available"


# noinspection PyCompatibility
@pytest.mark.parametrize(
    "package, resource, expected",
    [
        ("tests", "data/fispact/not_existing", "tests/data/fispact/not_existing"),
        ("mckit", "data/not_existing", "mckit/data/not_existing"),
    ],
)
def test_filename_resolver_when_resource_doesnt_exist(package, resource, expected):
    resolver = filename_resolver(package)
    actual = resolver(resource)
    assert not Path(actual).exists(), f"The resource {resource!r} should not be available"


def test_filename_resolver_when_package_doesnt_exist():
    resolver = filename_resolver("not_existing")
    with pytest.raises(ModuleNotFoundError):
        resolver("something.txt")


def test_path_resolver():
    resolver = path_resolver("tests")
    actual = resolver("utils/" + THIS_FILENAME)
    assert isinstance(actual, Path)
    assert actual.name == THIS_FILENAME
    assert actual.exists(), f"The file {THIS_FILENAME!r} should be available"


def test_path_resolver_in_own_package_with_separate_file():
    resolver = path_resolver("tests")
    assert resolver("__init__.py").exists(), "Should find '__init__.py' in the 'tests' package"


if __name__ == "__main__":
    pytest.main()



"""Code to estimate "closeness" of various objects."""
from typing import Callable, Iterable, Optional, Union

import itertools
import math

from abc import ABC, abstractmethod

import numpy as np

from mckit.constants import FLOAT_TOLERANCE
from numpy import ndarray


class MaybeClose(ABC):
    """Interface to be implemented by objects supporting the "closeness" estimation."""

    @abstractmethod
    def is_close_to(self, other: object, estimator: "EstimatorType") -> bool:
        """Objects can be estimated as close with some estimator."""


ComparableType = Optional[Union[Iterable, ndarray, float, int, MaybeClose]]
EstimatorType = Callable[[ComparableType, ComparableType], bool]


def tolerance_estimator(
    rtol: float = FLOAT_TOLERANCE,
    atol: float = FLOAT_TOLERANCE,
    equal_nan: bool = False,
) -> EstimatorType:
    """Estimates "closeness".

    For numpy arrays and float scalars uses math.isclose and numpy.allclose methods.
    For integers - direct comparison.
    Scans generic iterables and compares objects implementing MayBeClose interface.

    Returns:
        EstimatorType: estimator
    """

    def _estimate(a: ComparableType, b: ComparableType) -> bool:
        if isinstance(a, float) and isinstance(b, float):
            return math.isclose(a, b, rel_tol=rtol, abs_tol=atol)
        elif isinstance(a, ndarray) and isinstance(b, ndarray):
            return np.allclose(a, b, rtol, atol, equal_nan)
        elif issubclass(type(a), Iterable) and issubclass(type(b), Iterable):
            return all(_call(ai, bi) for ai, bi in itertools.zip_longest(a, b))
        elif isinstance(a, int) and isinstance(b, int):
            return a == b
        elif isinstance(a, MaybeClose) and isinstance(b, MaybeClose):
            return a.is_close_to(b, _call)
        else:
            raise TypeError(f"Not implemented for {type(a).__name__} and {type(b).__name__}")

    def _call(a: ComparableType, b: ComparableType) -> bool:
        if a is b:
            return True
        if a is None:
            return False  # `a` is None, but `b` is not None here
        if b is None:
            return True
            # TODO(dvp): in our use cases absent optional objects (like Transformations) are equal
        return _estimate(a, b)

    return _call


DEFAULT_TOLERANCE_ESTIMATOR = tolerance_estimator()



import inspect

from pathlib import Path

import pkg_resources as pkg


def filename_resolver(package=None):
    if package is None:
        module = inspect.getmodule(inspect.stack()[1][0])
        package = module.__name__

    resource_manager = pkg.ResourceManager()

    def func(resource):
        return resource_manager.resource_filename(package, resource)

    func.__doc__ = f"Computes file names for resources located in {package}"

    return func


def path_resolver(package=None):

    # The package is to be found here, otherwise inspect will find this package
    # instead of caller's package.
    if package is None:
        module = inspect.getmodule(inspect.stack()[1][0])
        package = module.__name__

    resolver = filename_resolver(package)

    def func(resource):
        filename = resolver(resource)
        return Path(filename)

    if package is None:
        package = "caller package"

    # noinspection PyCompatibility
    func.__doc__ = f"Computes Path for resources located in {package}"

    return func



# nopycln: file
from .cell_index import CellDummyIndex, CellNotFoundError, CellStrictIndex, DummyCell
from .composition_index import (
    CompositionDummyIndex,
    CompositionNotFoundError,
    CompositionStrictIndex,
    DummyComposition,
    DummyMaterial,
)
from .Lexer import Lexer
from .surface_index import DummySurface, SurfaceDummyIndex, SurfaceNotFoundError, SurfaceStrictIndex
from .transformation_index import (
    DummyTransformation,
    TransformationDummyIndex,
    TransformationNotFoundError,
    TransformationStrictIndex,
)
from .utils import *



# flake8: noca

from .check import check as do_check  # noca: F401
from .compose import compose as do_compose  # noca: F401
from .decompose import decompose as do_decompose  # noca: F401
from .split import split as do_split  # noca: F401
from .transform import transform as do_transform  # noca: F401



import re

from collections import deque
from itertools import product

import numpy as np


def read_mctal(filename, encoding="utf-8"):
    """Reads tally file.

    Parameters
    ----------
    filename : str
        Name of mctal file.
    encoding : str
        Name of encoding. Default: utf-8.

    Returns
    -------
    tallies : dict
        A dictionary of tally data.
    """
    with open(filename, encoding=encoding) as f:
        text = f.read()
    flags = re.MULTILINE + re.IGNORECASE
    header, *tally_texts = re.split("tally", text, flags=flags)
    tallies = {}
    for text in tally_texts:
        tally = parse_tally(text)
        tallies[tally["name"]] = tally
    return tallies


def parse_tally(text):
    """Parses text of tally."""
    tally = {}
    header_text, bin_text, val_text, tfc_text, comment = split_topics(text)

    tally["comment"] = comment.strip()
    tally.update(parse_tally_header(header_text))
    tally.update(parse_bins(bin_text))
    data, error = parse_values(val_text, tally["dims"])
    tally["data"] = data
    tally["error"] = error
    return tally


def parse_values(text, shape):
    tokens = deque(text.split())
    data = np.empty(shape)
    err = np.empty_like(data)
    for index in product(*map(range, shape)):
        data[index] = float(tokens.popleft())
        err[index] = float(tokens.popleft())
    return data, err


def split_topics(text):
    flags = re.MULTILINE + re.IGNORECASE
    text, tfc_text = re.split("tfc", text, maxsplit=1, flags=flags)
    text, val_text = re.split("vals", text, maxsplit=1, flags=flags)
    text, bin_text = re.split("^f", text, maxsplit=1, flags=flags)
    header_text, comment = re.split("\n", text, maxsplit=1)
    return header_text, bin_text, val_text, tfc_text, comment


PARTICLE_CODES = {1: "N", 2: "P", 4: "E"}
DETECTOR_TYPES = {0: "Non", 1: "Point", 2: "Ring", 3: "FIP", 4: "FIR", 5: "FIC"}


def parse_tally_header(text):
    """Parses text of tally header."""
    result = {}
    values = deque(text.split())
    result["name"] = int(values.popleft())
    p_code = int(values.popleft())
    result["particles"] = {par for code, par in PARTICLE_CODES.items() if code & p_code}
    result["type"] = DETECTOR_TYPES[int(values.popleft())]
    return result


def parse_bins(text):
    bin_texts = deque(re.split("^[a-z]", text, flags=re.M + re.I))
    result = {"dims": [], "bins": [], "vars": []}
    # f: cells, surfaces, detector bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), True, int)
    update_bins(result, dim_size, bin_values, "f")
    # d: detectors
    dim_size, bin_values = parse_bin(bin_texts.popleft(), False)
    update_bins(result, dim_size, bin_values, "d")
    # u: user bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), False)
    update_bins(result, dim_size, bin_values, "u")
    # s: segment bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), False)
    update_bins(result, dim_size, bin_values, "s")
    # m: multiplier bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), False)
    update_bins(result, dim_size, bin_values, "m")
    # c: cosine bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), True, float)
    update_bins(result, dim_size, bin_values, "c")
    # e: energy bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), True, float)
    update_bins(result, dim_size, bin_values, "e")
    # t: time bins
    dim_size, bin_values = parse_bin(bin_texts.popleft(), True, float)
    update_bins(result, dim_size, bin_values, "t")
    return result


def update_bins(result, dim_size, bin_values, var):
    if dim_size > 0:
        result["dims"].append(dim_size)
        result["bins"].append(bin_values)
        result["vars"].append(var)


def parse_bin(text, read_values=True, val_type=float):
    bin_header, values_list = text.split("\n", maxsplit=1)
    tokens = deque(bin_header.split())
    t = tokens.popleft()
    if t.isdigit():
        tokens.appendleft(t)
    dim_size = int(tokens.popleft())
    if dim_size == 1:
        dim_size = 0
    # if extra:
    #     tokens.popleft()
    tokens = deque(values_list.split())
    if read_values:
        bin_values = [val_type(t) for t in tokens]
    else:
        bin_values = list(range(dim_size))
    return dim_size, bin_values



# -*- coding: utf-8 -*-

"""Разложение текста модели на секции.

Читает модель, извлекает и раскладывает в указанную директорию отедельно файлы для ячеек, поверхностей,
материалов, трансформаций, sdef и прочие карты. Файлы соответственно: cells.txt, surfaces.txt,
materials.txt, transformations.txt, sdef.txt, cards.txt
"""
from typing import Iterable, Union

from pathlib import Path

import mckit.parser.mcnp_section_parser as sp

from mckit.cli.logging import logger

from ...constants import MCNP_ENCODING
from .common import check_if_path_exists

OUTER_LINE = "=" * 40
INNER_LINE = "-" * 40


def print_text(text: str, output_dir: Path, section_file_name: str, override: bool) -> None:
    if text:
        out = output_dir / section_file_name
        check_if_path_exists(out, override)
        out.write_text(text, encoding=MCNP_ENCODING)


def print_cards(
    cards: Iterable[sp.Card], output_dir: Path, section_file_name: str, override: bool
) -> None:
    if cards:
        out = output_dir / section_file_name
        check_if_path_exists(out, override)
        with out.open("w", encoding=MCNP_ENCODING) as fid:
            for card in cards:
                print(card.text, file=fid)


def split(
    output_dir: Path, mcnp_file_name: Union[str, Path], override: bool, separators=False
) -> None:
    logger.debug("Splitting model from {}", mcnp_file_name)
    if isinstance(mcnp_file_name, str):
        mcnp_file_name = Path(mcnp_file_name)
    assert output_dir.is_dir()
    with open(mcnp_file_name, encoding=MCNP_ENCODING) as fid:
        sections: sp.InputSections = sp.parse_sections(fid)
    print_text(sections.title, output_dir, "title.txt", override)
    print_cards(sections.cell_cards, output_dir, "cells.txt", override)
    print_cards(sections.surface_cards, output_dir, "surfaces.txt", override)
    if sections.data_cards:
        materials, transformations, sdef, tallies, others = sp.distribute_cards(sections.data_cards)
        print_cards(materials, output_dir, "materials.txt", override)
        print_cards(transformations, output_dir, "transformations.txt", override)
        print_cards(sdef, output_dir, "sdef.txt", override)
        print_cards(tallies, output_dir, "tallies.txt", override)
        print_cards(others, output_dir, "cards.txt", override)
    print_text(sections.remainder, output_dir, "remainder.txt", override)
    logger.debug("The parts of %s are saved to {}", mcnp_file_name, output_dir)
    if separators:
        write_separators(output_dir, mcnp_file_name.stem)


def write_separators(output: Path, model: str) -> None:
    for section in "cells surfaces materials transformations tallies".split():
        for start_end in "start end".split():
            if start_end == "start":
                first_line = OUTER_LINE
                second_line = INNER_LINE
            else:
                first_line = INNER_LINE
                second_line = OUTER_LINE
            text = (
                "c\n"
                + "c   "
                + first_line
                + "\n"
                + "c\n"
                + f"c   {start_end} of {model} {section}\n"
                + "c\n"
                + "c   "
                + second_line
                + "\n"
                + "c\n"
            )
            path: Path = output / f"{section}_{start_end}.txt"
            path.write_text(text, encoding=MCNP_ENCODING)
    path = output / "new_line.txt"
    path.write_text("\n")



"""Intercept log messages from the used libraries and pass them to `loguru`.

See https://github.com/Delgan/loguru
"""
from typing import Final

import logging
import sys

from os import environ

from loguru import logger

# class PropagateHandler(logging.Handler):
#     """Send events from loguru to standard logging"""
#     def emit(self, record):
#         logging.getLogger(record.name).handle(record)
#
#
# logger.add(PropagateHandler(), format="{message}")


class InterceptHandler(logging.Handler):
    """Send events from standard logging to loguru."""

    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


log = logging.getLogger()
log.addHandler(InterceptHandler())


MCKIT_CONSOLE_LOG_FORMAT: Final[str] = environ.get(
    "MCKIT_CONSOLE_LOG_FORMAT",
    default="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "
    "<level>{level: <8}</level> | "
    "<level>{message}</level>",
)


def init_logger(logfile, quiet, verbose, *, stderr_format: str = MCKIT_CONSOLE_LOG_FORMAT):
    stderr_level: str = "INFO"
    if quiet:
        stderr_level = "WARNING"
    elif verbose:
        stderr_level = "TRACE"
    logger.remove()
    if stderr_format:
        logger.add(
            sys.stderr,
            format=stderr_format,
            level=stderr_level,
            backtrace=False,
            diagnose=False,
        )
    if logfile:
        logger.add(logfile, rotation="100 MB", level="TRACE")



"""Features, common for all cards."""
from __future__ import annotations

from typing import Any, Dict, List, Optional, Text, cast

from abc import ABC, abstractmethod
from functools import reduce
from operator import xor

from mckit.utils.named import Name

from .printer import print_card


class Card(ABC):
    """Features, common for all cards."""

    def __init__(self, **options: Dict[str, Any]):
        self.options: Dict[str, Any] = options

    def __str__(self):
        # TODO dvp: option `name` is printed twice,
        #           (second time as option)
        #           This should be explicit property of this class instance
        return f'{self.name()}: "{self.options}"'

    @property
    def is_anonymous(self) -> bool:
        """Is the card is named?"""
        return self.name() is None

    @property
    def has_original(self) -> bool:
        """Has original text stored in options."""
        return "original" in self.options

    @property
    def original(self) -> Optional[str]:
        """Original text from an MCNP model."""
        return cast(Optional[str], self.options.get("original", None))

    @property
    def has_comment_above(self) -> bool:
        """Has comment above stored in options."""
        return "comment_above" in self.options

    @property
    def comment_above(self) -> Optional[str]:
        """Comment located above this card in an MCNP model."""
        return cast(Optional[str], self.options.get("comment_above", None))

    def name(
        self,
    ) -> Optional[
        Name
    ]:  # TODO dvp: we'd better have special property name, don't use options for that
        """Returns card's name."""
        return self.options.get("name", None)

    def rename(self, new_name) -> "Card":
        """Renames the card."""
        self.options["name"] = new_name
        self.drop_original()
        return self

    @abstractmethod
    def mcnp_words(self, pretty=False) -> List[Text]:
        """Gets a list of card words."""

    def mcnp_repr(self, pretty: bool = False) -> str:
        """Gets str representation of the card."""
        # TODO dvp: try to use original texts, if available - this will preserve comments
        return cast(str, print_card(self.mcnp_words(pretty)))

    def drop_original(self) -> None:
        """Drop original text, if any.

        Do this, if the card is changed and doesn't correspond to original text anymore.
        """
        if "original" in self.options:
            del self.options["original"]

    def add_comment(self, *comment: str) -> None:
        """Add a comment to this card."""
        self.options.setdefault("comment", []).extend(comment)

    def __hash__(self) -> int:
        return reduce(xor, (hash(k) ^ hash(v) for k, v in self.options.items()), 0)

    def __eq__(self, other) -> bool:
        return self is other or self.options == other.options



from typing import List

from io import StringIO

import mckit.parser.mcnp_section_parser as sp
import pytest

from mckit.parser.mcnp_section_parser import Card, Kind


@pytest.mark.parametrize(
    "text,expected",
    [
        ("aaa\n\nbbb", ["aaa", "bbb"]),
        ("aaa\n   \nbbb", ["aaa", "bbb"]),
        ("aaa\nbbb", ["aaa\nbbb"]),
    ],
)
def test_blank_line_pattern(text: str, expected: List[str]) -> None:
    actual = sp.BLANK_LINE_PATTERN.split(text)
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        ("1 0 1\n", ("1 0 1\n", 0)),
        ("1 0 1 $bla bla bla\n", ("1 0 1\n", 1)),
        (
            "1 0 1 $bla bla bla\n" + "c the comment with the space before c\n" + "   2 -3",
            ("1 0 1\n" + "   2 -3", 2),
        ),
    ],
)
def test_removing_comment_pattern(text, expected):
    actual = sp.REMOVE_COMMENT_PATTERN.subn("", text)
    assert expected == actual


@pytest.mark.parametrize(
    "text,comment,card",
    [
        ("1 0 1\n", None, "1 0 1\n"),
        (
            """\
c the preceding comment
1 0 1""",
            """\
c the preceding comment
""",
            "1 0 1",
        ),
        (
            """\
c the preceding comment
1 0 1
 c The next card comment (with space before it)
  2 $next card""",
            """\
c the preceding comment
""",
            "1 0 1\n",
        ),
        (
            """\
C ----------------------------------------------------------------------------C
C    TRANSFORMATIONS                                                          C
C    ROTATION WITH RESPECT TO Z-AXIS (-20,|5|,20)                             C
C ----------------------------------------------------------------------------C
*TR1    0  0  0
        20.0000    70.0000  90
       110.0000    20.0000  90
       90        90         0""",
            """\
C ----------------------------------------------------------------------------C
C    TRANSFORMATIONS                                                          C
C    ROTATION WITH RESPECT TO Z-AXIS (-20,|5|,20)                             C
C ----------------------------------------------------------------------------C
""",
            """\
*TR1    0  0  0
        20.0000    70.0000  90
       110.0000    20.0000  90
       90        90         0""",
        ),
    ],
)
def test_card_pattern(text, comment, card):
    res = sp.CARD_PATTERN.match(text)
    groups = res.groupdict()
    actual_comment = groups["comment"]
    assert comment == actual_comment
    actual_card = groups["card"]
    assert actual_card == card


@pytest.mark.parametrize(
    "text,expected,tag,number",
    [
        ("f1", "f1", "f", 1),
        ("FMESH10", "FMESH10", "FMESH", 10),
        ("fm999", "fm999", "fm", 999),
        ("fc1", "fc1", "fc", 1),
        ("de1", "de1", "de", 1),
        ("df1", "df1", "df", 1),
    ],
)
def test_tally_pattern(text: str, expected: str, tag: str, number: int) -> None:
    res = sp.TALLY_PATTERN.fullmatch(text)
    actual = res.group(0) if res else None
    assert actual == expected, f"Should match pattern '{text}'"
    actual_tag = res["tag"]
    assert actual_tag == tag, "Tags differ"
    actual_number = int(res["number"])
    assert actual_number == number, "Numbers differ"


@pytest.mark.parametrize("text", ["f", "111"])
def test_tally_pattern_bad_path(text: str) -> None:
    res = sp.TALLY_PATTERN.fullmatch(text)
    assert res is None, f"Should not match pattern '{text}'"


@pytest.mark.parametrize(
    "text,cards,kind",
    [
        (
            """
c the preceding comment
1 0 1
c next comment
  2 $next card (starts with less than 5 spaces)
"""[
                1:-1
            ],
            [
                Card("c the preceding comment"),
                Card("1 0 1", kind=Kind.CELL),
                Card("c next comment"),
                Card("  2 $next card (starts with less than 5 spaces)", kind=Kind.CELL),
            ],
            Kind.CELL,
        ),
        (
            """
c the preceding comment
1 0 1
c inner comment
     2 $continuation
c the second preceding comment
2 0 -1 $the next card
c the trailing comment
c the second line of the trailing comment
"""[
                1:-1
            ],
            [
                Card("c the preceding comment"),
                Card(
                    """
1 0 1
c inner comment
     2 $continuation
"""[
                        1:-1
                    ],
                    kind=Kind.CELL,
                ),
                Card("c the second preceding comment"),
                Card("2 0 -1 $the next card", kind=Kind.CELL),
                Card(
                    """
c the trailing comment
c the second line of the trailing comment
"""[
                        1:-1
                    ]
                ),
            ],
            Kind.CELL,
        ),
        (
            """
  cut 5j  $ card starts in column < 5
ctme 3000
"""[
                1:
            ],
            [Card("  cut 5j  $ card starts in column < 5"), Card("ctme 3000")],
            None,
        ),
        (
            """
m100
      1001.31c 0.6666
      8000.21c 0.3334
"""[
                1:-1
            ],
            [
                Card(
                    """
m100
      1001.31c 0.6666
      8000.21c 0.3334
"""[
                        1:-1
                    ],
                    kind=Kind.MATERIAL,
                )
            ],
            None,
        ),
    ],
)
def test_split_to_cards(text, cards, kind):
    actual_cards = list(sp.split_to_cards(text, kind))
    assert actual_cards == cards


@pytest.mark.parametrize(
    "text",
    [
        (
            """
sdef
sp1
si1
ds3
sb45
wwp:n
"""[
                1:-1
            ]
        )
    ],
)
def test_sdef_cards(text):
    actual_cards = list(sp.split_to_cards(text))
    for card in actual_cards:
        assert card.is_sdef, "Should be SDEF card"


def test_card_constructor():
    description = "1 0 1"
    c = Card(description)
    assert c.text == description


def test_input_sections_constructor():
    title = "Testing"
    cell_cards = [Card("1 0 1"), Card("2 0 -1")]
    surface_cards = [Card("1 so 100")]
    t = sp.InputSections(title, cell_cards, surface_cards, [Card("sdef")])
    assert t.title == title


@pytest.mark.parametrize(
    "text,expected,kind",
    [
        (
            """
1 0 1 $bla bla bla
"""[
                1:-1
            ],
            [Card("1 0 1", kind=Kind.CELL)],
            Kind.CELL,
        ),
        (
            """
c some comment
c second line
1 0 1 $bla bla bla
     2 $continuation
c trailing comment
"""[
                1:-1
            ],
            [
                Card(
                    """
1 0 1 2
"""[
                        1:-1
                    ],
                    kind=Kind.CELL,
                )
            ],
            Kind.CELL,
        ),
        (
            """
c some comment
c second line
1 0 1 $bla bla bla
     2 $continuation
c trailing comment
2
     0 -1  $something
c z-z-zz-z-z-z
"""[
                1:-1
            ],
            [
                Card(
                    """
1 0 1 2
"""[
                        1:-1
                    ],
                    kind=Kind.CELL,
                ),
                Card(
                    """
2 0 -1
"""[
                        1:-1
                    ],
                    kind=Kind.CELL,
                ),
            ],
            Kind.CELL,
        ),
    ],
)
def test_clean_mcnp_cards(text, expected, kind):
    actual = list(sp.clean_mcnp_cards(sp.split_to_cards(text, kind)))
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        (
            """
test
1 0 1 $bla bla bla

1 so 1

sdef
"""[
                1:
            ],
            None,
        ),
        (
            """
test
c some comment
c second line
1 0 1 $bla bla bla
     2 $continuation
2 0 -1
c trailing comment

1 so 1

sdef
"""[
                1:
            ],
            None,
        ),
        (
            """
test
c some comment
c second line
1 0 1 $bla bla bla
     2 $continuation
c trailing comment
2
    0 -1  $something
c z-z-zz-z-z-z

1 so 1

sdef
"""[
                1:
            ],
            None,
        ),
        (
            """
continue
ctme 3000
"""[
                1:
            ],
            None,
        ),
    ],
)
def test_print(text, expected):
    stream = StringIO(text)
    sections = sp.parse_sections(stream)
    out = StringIO()
    sections.print(out)
    if expected is None:
        expected = text.strip()
    actual = out.getvalue().strip()
    assert actual == expected



"""Utility code to use in all other modules."""
from .accept import TVisitor, accept, on_unknown_acceptor
from .io import MCNP_ENCODING, check_if_all_paths_exist, check_if_path_exists, make_dir, make_dirs
from .misc import (
    MAX_DIGITS,
    are_equal,
    compute_hash,
    deepcopy,
    filter_dict,
    get_decades,
    is_in,
    is_sorted,
    make_hashable,
    mids,
    prettify_float,
    round_array,
    round_scalar,
    significant_array,
    significant_digits,
)
from .resource import filename_resolver, path_resolver
from .tolerance import FLOAT_TOLERANCE



from __future__ import annotations

from typing import Generator, List

import os
import sys
import sysconfig

from ctypes import cdll
from logging import getLogger
from pathlib import Path

_LOG = getLogger(__name__)

HERE = Path(__file__).parent.absolute()
WIN = sys.platform.startswith("win32") and "mingw" not in sysconfig.get_platform()
MACOS = sys.platform.startswith("darwin")


def _library_base_name(_lib_name: str) -> str:
    return _lib_name if WIN else "lib" + _lib_name


SUFFIX = ".dll" if WIN else ".dylib" if MACOS else ".so"

if WIN or MACOS:

    def _combine_version_and_suffix(version: int, suffix: str) -> str:
        return f".{version}{suffix}"  # .2.dll or .2.dylib

else:  # Linux

    def _combine_version_and_suffix(version: int, suffix: str) -> str:
        return f"{suffix}.{version}"  # .so.2


def _iterate_suffixes_with_version(max_version: int = 2) -> Generator["str", None, None]:
    while max_version >= 0:
        yield _combine_version_and_suffix(max_version, SUFFIX)
        max_version -= 1
    yield SUFFIX


SHARED_LIBRARY_DIRECTORIES: List[Path] = []

if WIN:
    SHARED_LIBRARY_DIRECTORIES.append(Path(sys.prefix, "Library", "bin"))
else:
    ld_library_path = os.environ.get("DYLD_LIBRARY_PATH" if MACOS else "LD_LIBRARY_PATH")
    if ld_library_path:
        SHARED_LIBRARY_DIRECTORIES.extend(map(Path, ld_library_path.split(":")))
    SHARED_LIBRARY_DIRECTORIES.append(Path(sys.prefix, "lib"))

SHARED_LIBRARY_DIRECTORIES.append(HERE)


def _preload_library(lib_name: str, max_version: int = 2) -> None:
    for d in SHARED_LIBRARY_DIRECTORIES:
        for s in _iterate_suffixes_with_version(max_version):
            p = Path(d, _library_base_name(lib_name)).with_suffix(s)
            if p.exists():
                cdll.LoadLibrary(str(p))
                _LOG.info("Found library: {}", p.absolute())
                return
    raise EnvironmentError(f"Cannot preload library {lib_name}")


def _init():
    if WIN and hasattr(os, "add_dll_directory"):  # Python 3.7 doesn't have this method
        for _dir in SHARED_LIBRARY_DIRECTORIES:
            os.add_dll_directory(str(_dir))
    _preload_library("mkl_rt", max_version=2)
    _preload_library("nlopt", max_version=0)


_init()

# We have to import these libraries here, after preloading libraries.

import mckit.geometry as geometry

from mckit.body import Body, Shape
from mckit.surface import Cone, Cylinder, GQuadratic, Plane, Sphere, Torus, create_surface



"""Code to load source distribution and print corresponding MCNP SDEF."""
from __future__ import annotations

from typing import Dict, List, Optional, Set, Tuple, Union

from functools import reduce

import numpy as np

from .printer import print_card, separate


class Distribution:
    """Represents a distribution of source variable.

    If the len of probs equals to the len of values, then the distribution is
    either of discrete variable (if is_discrete flag is set) or probs sets
    probability density values. If len(probs) + 1 == len(values), then
    values are bin boundaries for continuous distribution.

    Parameters
    ----------
    name : int
        Distribution's name. It is needed for MCNP.
    values : list
        A list of values, that variable can be equal to. Values can be a float,
        an int or other Distribution instances.
    probs : list or Distribution
        A list of probabilities or a distribution this one depends on.
        Probabilities need not be normalized.
    variable : str
        Source variable's name. Default: None.
    is_discrete : bool
        Indicate that the variable is discrete. Default: False.

    Methods
    -------
    mcnp_repr()
        Returns a string representation of corresponding MCNP card.
    get_inner()
        Gets all distributions this distribution depends on.
    depends_on()
        Gets distribution this one depends on.

    Properties
    ----------
    name : int
        Distribution's name.
    size : int
        Size of the distribution. It is the number of values or bins.
    variable : str
        Name of source variable.
    """

    def __init__(
        self,
        name: int,
        values: list[Union[int, float, "Distribution"]],
        probs: Union[List[float], "Distribution"],
        variable: Optional[str] = None,
        is_discrete: bool = False,
    ) -> None:
        self._name = name
        self._var = variable
        Distribution.check_distr(values, len(probs), is_discrete)
        self._is_discrete = is_discrete
        self._values = list(values)
        if isinstance(probs, Distribution):
            self._probs: Union[List[float], "Distribution"] = probs
        else:
            self._probs = list(probs)
        if (
            isinstance(self._probs, list)
            and len(self._probs) == len(values)
            and not is_discrete
            and not isinstance(self._values[0], Distribution)
        ):
            self._is_pdf = True
        else:
            self._is_pdf = False

    @property
    def name(self) -> int:
        """Gets distribution's name."""
        return self._name

    @property
    def size(self) -> int:
        """Gets distribution's size."""
        return len(self._probs)

    @property
    def variable(self) -> Optional[str]:
        """Gets distribution's source variable name."""
        return self._var

    def mcnp_repr(self) -> str:
        """Returns a string representation of corresponding MCNP card."""
        tokens = []
        discrete = self._is_discrete
        is_pdf = self._is_pdf
        if isinstance(self._values[0], Distribution):
            tokens.append("S")
            for v in self._values:
                tokens.append(str(v.name))
        else:
            if isinstance(self._values[0], list):
                values = reduce(list.__add__, self._values)
            else:
                values = self._values
            if discrete:
                tokens.append("L")
            elif is_pdf:
                tokens.append("A")
            else:
                tokens.append("H")
            for v in values:
                tokens.append(str(v))

        if isinstance(self._probs, Distribution):
            tokens.insert(0, f"DS{self._name}")
            prob_tokens = None
        else:
            tokens.insert(0, f"SI{self._name}")
            prob_tokens = [f"SP{self._name}"]
            if not is_pdf:
                prob_tokens.append("D")
            if len(self._values) == len(self._probs) + 1:
                prob_tokens.append("0")
            for p in self._probs:
                prob_tokens.append(str(p))
        card = print_card(separate(tokens))
        if prob_tokens:
            card += "\n" + print_card(separate(prob_tokens))
        return card

    def get_inner(self) -> Set["Distribution"]:
        """Gets nested distributions this one depends on.

        If values of this distribution are distributions themselves,
        then they are returned.

        Returns:
            Set:  A set of nested distributions.

        Raises:
            TypeError: when one of the values is not a `Distribution`.
        """
        dists: Set["Distribution"] = set()
        if isinstance(self._values[0], Distribution):
            for v in self._values:
                if not isinstance(v, Distribution):
                    raise TypeError(f"Expect `Distribution` type, actual {type(v)}")
                dists.add(v)
        return dists

    def __len__(self) -> int:
        """Size of distribution."""
        return len(self._probs)

    def depends_on(self) -> Optional["Distribution"]:
        """Gets distribution this one depends on.

        Returns:
            Distribution|None:  Distribution this one depends on, if any. None otherwise.
        """
        if isinstance(self._probs, Distribution):
            return self._probs
        else:
            return None

    @staticmethod
    def check_distr(
        values: list[Union[int, float, "Distribution"]], size: int, is_discrete: bool
    ) -> None:
        """Checks if the distribution is correct.

        Args:
            values: List of variable values.
            size: The length of intensity matrix along variable dimension.
            is_discrete:  True <=> the distribution is discrete.

        Raises:
            ValueError: if the distribution is neither discrete, nor pdf, nor histogram.
        """
        discr_or_pdf = len(values) == size
        histogram = len(values) == size + 1 and not is_discrete
        if not discr_or_pdf and not histogram:
            raise ValueError("Inconsistent size of values.")


def create_bin_distributions(
    bins: list[float], start_name: int = 1
) -> Tuple[int, list["Distribution"]]:
    """Creates bin distributions for specified bins.

    A list of distributions created. Index in the list corresponds to the  index of bin.

    Args:
        bins:   A list of bin boundaries.
        start_name:  Starting name for distributions.
                     For every new distribution the name is incremented by 1.

    Returns:
        free_name, list of distributions
    """
    distributions = []
    for low, high in zip(bins[:-1], bins[1:]):
        distributions.append(Distribution(start_name, [low, high], [1]))
        start_name += 1
    return start_name, distributions


def expand_matrix_distribution(
    intensities: np.ndarray,
    *var_values: list[float],
    start_name: int = 1,
) -> Tuple[int, Tuple[list[float], ...], List[float]]:
    """Converts matrix distribution to the len(var_values) linear.

    Creates:
        - expr_valu_values -  A tuple of lists of values for each source variable.
        - exp_intensities -   A list of expanded intensities for every hyperbin.

    Args:
        intensities :  A matrix of source intensities.
        var_values :   A tuple of variable values along each axis.
        start_name :   Starting name for distributions. Default: 1.

    Returns:
        free_name: int,  exp_var_values : tuple, exp_intensities : list[float]

    Raises:
        ValueError: if the distribution is neither discrete, nor pdf, nor histogram.
    """
    if len(intensities.shape) != len(var_values):
        msg = "Length of var_values must be equal to the number of intensities dimensions."
        raise ValueError(msg)
    uniq_values = []
    exp_var_values = []
    for dim, values in zip(intensities.shape, var_values):
        if Distribution.check_distr(values, dim):
            uniq_values.append(values)
        else:
            start_name, dists = create_bin_distributions(values, start_name)
            uniq_values.append(dists)
        exp_var_values.append([])

    exp_intensities = []
    for index, intensity in intensities:
        for j, i in enumerate(index):
            exp_var_values[j].append(uniq_values[j][i])
        exp_intensities.append(intensity)
    return start_name, tuple(exp_var_values), exp_intensities


class Source:
    """Represents particle source.

    Parameters
    ----------
    variables : dict
        A dictionary of source variables. var_name -> var_value. The value
        can be either float (int) value or Distribution instance.

    Methods
    -------
    mcnp_repr()
        Gets a string representation of corresponding MCNP card.
    """

    def __init__(self, **variables: Dict[int, Union[int, float, Distribution]]) -> None:
        self._variables = variables

    def mcnp_repr(self) -> str:
        """Gets a string representation of corresponding MCNP card."""
        tokens = ["SDEF"]
        cards = []
        extra_cards = []

        def _var_repr(value) -> str:
            if isinstance(value, Distribution):
                dep = value.depends_on()
                result = f"D{value.name}"
                if dep:
                    result = f"F{dep.variable} " + result
            else:
                result = str(value)
            return result

        for k, v in self._variables.items():
            tokens.append(f"{k}={_var_repr(v)}")
            if isinstance(v, Distribution):
                cards.append(v.mcnp_repr())
                for ec in sorted(v.get_inner(), key=lambda x: x.name):
                    extra_cards.append(ec.mcnp_repr())
        cards.insert(0, print_card(separate(tokens)))
        cards.extend(extra_cards)
        return "\n".join(cards)



from typing import Dict, List, Set, Union

import tempfile
import textwrap

from copy import deepcopy

import numpy as np
import pytest

from mckit.body import Body, Card, Shape
from mckit.box import Box
from mckit.material import Composition, Element, Material
from mckit.parser import ParseResult, from_file, from_text
from mckit.surface import Sphere, Surface, create_surface
from mckit.transformation import Transformation
from mckit.universe import (
    NameClashError,
    Universe,
    cell_selector,
    collect_transformations,
    surface_selector,
)
from mckit.utils.resource import filename_resolver

data_filename_resolver = filename_resolver("tests")

TStatItem = Dict[
    int, Union[List[int], Set[Universe]]
]  # TODO dvp: cool but isn't this too much freedom?
TStat = Dict[str, TStatItem]


@pytest.fixture(scope="module")
def universe():
    cases = {
        1: "universe_test_data/universe1.i",
        2: "universe_test_data/universe2.i",
        3: "universe_test_data/universe3.i",
        4: "universe_test_data/universe4.i",
        5: "universe_test_data/universe5.i",
        1002: "universe_test_data/universe1002.i",
        1012: "universe_test_data/universe1012.i",
        1022: "universe_test_data/universe1022.i",
    }

    def _universe(case: int) -> Universe:
        result: ParseResult = from_file(data_filename_resolver(cases[case]))
        return result.universe

    return _universe


@pytest.mark.parametrize(
    "case, points, answer",
    [
        (
            1,
            [[-2, 0, 4], [10, 10, -10], [0, 0, 0], [0, 0, -2.5], [0, 1, 0]],
            [1, 3, 0, 2, 0],
        )
    ],
)
def test_points(universe, case, points, answer):
    u = universe(case)
    result = u.test_points(points)
    np.testing.assert_array_equal(result, answer)


_emp = Shape("R")


@pytest.mark.parametrize(
    "kwargs",
    [
        {"name": 3, "verbose_name": "verbose", "comment": "comment"},
        {},
        {"name": 4, "comment": ["1", "2"]},
    ],
)
@pytest.mark.parametrize(
    "cells", [[Body(_emp, name=2), Body(_emp, name=4)], [], [Body(_emp, name=5)]]
)
def test_init(cells, kwargs):
    u = Universe(cells, **kwargs)
    assert len(u._cells) == len(cells)
    for c1, c2 in zip(u._cells, cells):
        assert c1.name() == c2.name()
        assert c1.shape == c2.shape
    assert u.name() == kwargs.get("name", 0)
    assert u.verbose_name == kwargs.get("verbose_name", str(kwargs.get("name", 0)))
    assert u._comment == kwargs.get("comment", None)


@pytest.mark.parametrize(
    "case, cells, name_rule, new_name, new_surfs, new_comps",
    [
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=2),
            "keep",
            None,
            [],
            [],
        ),  # 0
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=2)), name=8),
            "keep",
            [8],
            None,
            [],
        ),  # 1
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=7),
            "keep",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),
        # 2
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=2),
            "clash",
            [5],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),
        # 3
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=2)), name=8),
            "clash",
            [8],
            [Sphere([0, 3, 0], 0.5, name=6)],
            [],
        ),
        # 4
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=7),
            "clash",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),
        # 5
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=2),
            "new",
            [5],
            [Sphere([0, 3, 0], 0.5, name=6)],
            [],
        ),
        # 6
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=2)), name=8),
            "new",
            [5],
            [Sphere([0, 3, 0], 0.5, name=6)],
            [],
        ),
        # 7
        (
            1,
            Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=7),
            "new",
            [5],
            [Sphere([0, 3, 0], 0.5, name=6)],
            [],
        ),
        # 8
        (
            1,
            Body(Shape("C", Sphere([0, 0, 0], 2, name=8)), name=2),
            "new",
            [5],
            [],
            [],
        ),  # 9
        (
            1,
            Body(Shape("C", Sphere([0, 0, 0], 2, name=2)), name=8),
            "keep",
            [8],
            [],
            [],
        ),  # 10
        (
            1,
            Body(Shape("C", Sphere([0, 0, 0], 2, name=8)), name=7),
            "new",
            [5],
            [],
            [],
        ),  # 11
        (
            1,
            [
                Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=8),
                Body(Shape("C", Sphere([0, 4, 0], 0.5, name=8)), name=9),
            ],
            "keep",
            [8, 9],
            None,
            [],
        ),  # 12
        (
            1,
            [
                Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=9),
                Body(Shape("C", Sphere([0, 4, 0], 0.5, name=9)), name=9),
            ],
            "keep",
            None,
            [8, 9],
            [],
        ),  # 13
        (
            1,
            [
                Body(Shape("C", Sphere([0, 3, 0], 0.5, name=2)), name=2),
                Body(Shape("C", Sphere([0, 4, 0], 0.5, name=2)), name=2),
            ],
            "clash",
            [5, 6],
            [Sphere([0, 3, 0], 0.5, name=6), Sphere([0, 4, 0], 0.5, name=7)],
            [],
        ),  # 14
        (
            1,
            [
                Body(Shape("C", Sphere([0, 3, 0], 0.5, name=8)), name=9),
                Body(Shape("C", Sphere([0, 4, 0], 0.5, name=9)), name=9),
            ],
            "new",
            [5, 6],
            [Sphere([0, 3, 0], 0.5, name=6), Sphere([0, 4, 0], 0.5, name=7)],
            [],
        ),  # 15
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(
                    composition=Composition(atomic=[("C-12", 1)], name=2, lib="31c"),
                    density=2.0,
                ),
            ),
            "keep",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),  # 16
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(
                    composition=Composition(atomic=[("C-12", 1)], name=1, lib="31c"),
                    density=2.0,
                ),
            ),
            "keep",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),  # 17
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(
                    composition=Composition(atomic=[("C-12", 1)], name=3, lib="31c"),
                    density=2.0,
                ),
            ),
            "keep",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [],
        ),  # 18
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(composition=Composition(atomic=[("Fe-56", 1)], name=2), density=2.0),
            ),
            "keep",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            None,
        ),  # 19
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(composition=Composition(atomic=[("Fe-56", 1)], name=2), density=2.0),
            ),
            "new",
            [5],
            [Sphere([0, 3, 0], 0.5, name=6)],
            [Composition(atomic=[("Fe-56", 1)], name=3)],
        ),  # 20
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(composition=Composition(atomic=[("Fe-56", 1)], name=2), density=2.0),
            ),
            "clash",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [Composition(atomic=[("Fe-56", 1)], name=3)],
        ),  # 21
        (
            1,
            Body(
                Shape("C", Sphere([0, 3, 0], 0.5, name=8)),
                name=7,
                MAT=Material(composition=Composition(atomic=[("Fe-56", 1)], name=6), density=2.0),
            ),
            "clash",
            [7],
            [Sphere([0, 3, 0], 0.5, name=8)],
            [Composition(atomic=[("Fe-56", 1)], name=6)],
        ),  # 22
    ],
)
def test_add_cells(universe, case, cells, name_rule, new_name, new_surfs, new_comps):
    u = universe(case)
    s_before = u.get_surfaces()
    c_before = u.get_compositions()
    if new_surfs is None or new_name is None or new_comps is None:
        with pytest.raises(NameClashError):
            u.add_cells(cells, name_rule=name_rule)
    else:
        u.add_cells(cells, name_rule=name_rule)
        s_after = u.get_surfaces()
        c_after = u.get_compositions()
        added_cells = u._cells[-len(new_name) :]
        if isinstance(cells, Body):
            cells = [cells]
        for added_cell, cell, name in zip(added_cells, cells, new_name):
            assert added_cell.shape == cell.shape
            assert added_cell is not cell
            assert added_cell.name() == name
            assert added_cell.options["U"] is u
            if cell.material() is not None:
                assert added_cell.material().composition == cell.material().composition

        assert_change(s_before, s_after, new_surfs)
        assert_change(c_before, c_after, new_comps)


@pytest.mark.parametrize(
    "case, cells, shapes",
    [
        (
            1,
            Body(Shape("C", create_surface("P", 0, 0, -1, -3, name=8)), name=5),
            [Shape("S", create_surface("P", 0, 0, 1, 3, name=2))],
        )
    ],
)
def test_add_cells_neg(universe, case, cells, shapes):
    u = universe(case)
    s_before = u.get_surfaces()
    u.add_cells(cells, name_rule="keep")
    s_after = u.get_surfaces()
    if isinstance(cells, Body):
        cells = [cells]
    added_cells = u._cells[-len(cells) :]
    if isinstance(cells, Body):
        cells = [cells]
    for added_cell, cell, shape in zip(added_cells, cells, shapes):
        assert added_cell.shape == shape
        assert added_cell is not cell
        assert added_cell.name() == cell.name()
        assert added_cell.options["U"] is u
        if cell.material() is not None:
            assert added_cell.material().composition == cell.material().composition

    assert_change(s_before, s_after, {})


@pytest.mark.parametrize(
    "case, common_materials, ans_compositions",
    [
        (
            1,
            set(),
            {
                Composition(atomic=[("C-12", 1)], name=1),
                Composition(atomic=[("H-1", 2), ("O-16", 1)], name=2, lib="31c"),
            },
        ),
        (
            1,
            {Composition(atomic=[("C-12", 1)], name=10, lib="31c")},
            {
                Composition(atomic=[("C-12", 1)], name=10, lib="31c"),
                Composition(atomic=[("H-1", 2), ("O-16", 1)], name=2, lib="31c"),
            },
        ),
        (
            1,
            {
                Composition(atomic=[("C-12", 1)], name=10, lib="31c"),
                Composition(atomic=[("H-1", 2), ("O-16", 1)], name=11, lib="31c"),
            },
            {
                Composition(atomic=[("C-12", 1)], name=10, lib="31c"),
                Composition(atomic=[("H-1", 2), ("O-16", 1)], name=11, lib="31c"),
            },
        ),
        (
            1,
            {Composition(atomic=[("Fe-56", 1)], name=10)},
            {
                Composition(atomic=[("C-12", 1)], name=1, lib="31c"),
                Composition(atomic=[("H-1", 2), ("O-16", 1)], name=2),
            },
        ),
        (1, {Composition(atomic=[("Fe-56", 1)], name=1, lib="31c")}, None),
    ],
)
def test_common_materials(universe, case, common_materials, ans_compositions):
    u = universe(case)
    if ans_compositions is None:
        with pytest.raises(NameClashError):
            Universe(u, common_materials=common_materials)
    else:
        new_u = Universe(u, common_materials=common_materials)
        new_comps = new_u.get_compositions()
        assert new_comps == ans_compositions
        common = {c.name(): c for c in common_materials}
        for c in new_comps:
            if c in common_materials:
                assert c is common[c.name()]


def assert_change(before, after, new_items):
    for s in before:
        assert s in after
    assert len(after) - len(before) == len(new_items)
    diff = after.difference(before)
    assert diff == set(new_items)
    assert {s.name() for s in diff} == {s.name() for s in new_items}


@pytest.mark.parametrize(
    "case, recursive, answer_data",
    [
        (
            1,
            False,
            [
                (1, "SO", [2]),
                (2, "PZ", [3]),
                (3, "PZ", [5]),
                (4, "C/Z", [-2, 0, 1]),
                (5, "SZ", [3.5, 10]),
            ],
        ),
        (
            1,
            True,
            [
                (1, "SO", [2]),
                (2, "PZ", [3]),
                (3, "PZ", [5]),
                (4, "C/Z", [-2, 0, 1]),
                (5, "SZ", [3.5, 10]),
            ],
        ),
    ],
)
def test_get_surfaces(universe, case, recursive, answer_data):
    answer = {create_surface(k, *p, name=n) for n, k, p in answer_data}
    u = universe(case)
    surfaces = u.get_surfaces(inner=recursive)
    assert surfaces == answer
    names_ans = {x[0] for x in answer_data}
    names = {s.name() for s in surfaces}
    assert names == names_ans


@pytest.mark.parametrize(
    "case, answer", [(5, {Composition(atomic=[("6012", 1)], name=10)}), (1, set())]
)
def test_find_common_materials(universe, case, answer):
    u = universe(case)
    cm = u._common_materials
    assert cm == answer
    assert {c.name() for c in cm} == {c.name() for c in answer}


@pytest.mark.parametrize(
    "case, answer",
    [
        (
            1,
            {
                Composition(atomic=[(Element("C-12", lib="31c"), 1)], name=1),
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                ),
            },
        ),
        (2, set()),
        (
            3,
            {
                Composition(atomic=[(Element("C-12", lib="31c"), 1)], name=4),
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=3,
                ),
            },
        ),
    ],
)
def test_get_compositions(universe, case, answer):
    u = universe(case)
    comps = u.get_compositions()
    assert comps == answer
    names_ans = {c.name() for c in answer}
    names = {c.name() for c in comps}
    assert names == names_ans


@pytest.mark.parametrize(
    "tr",
    [
        Transformation(translation=[-3, 2, 0.5]),
        Transformation(translation=[1, 2, 3]),
        Transformation(translation=[-4, 2, -3]),
        Transformation(
            translation=[3, 0, 9],
            rotation=[30, 60, 90, 120, 30, 90, 90, 90, 0],
            indegrees=True,
        ),
        Transformation(
            translation=[1, 4, -2],
            rotation=[0, 90, 90, 90, 30, 60, 90, 120, 30],
            indegrees=True,
        ),
        Transformation(
            translation=[-2, 5, 3],
            rotation=[30, 90, 60, 90, 0, 90, 120, 90, 30],
            indegrees=True,
        ),
    ],
)
@pytest.mark.parametrize("case, extent", [(1, [20, 10, 10])])
def test_transform(universe, case, tr, extent):
    u = universe(case)
    points = np.random.random((50000, 3))
    points -= np.array([0.5, 0.5, 0.5])
    points *= np.array(extent)
    test_results = [c.shape.test_points(points) for c in u]

    tr_u = u.transform(tr)
    tr_points = tr.apply2point(points)
    tr_results = [c.shape.test_points(tr_points) for c in tr_u]
    assert len(u._cells) == len(tr_u._cells)
    for r, r_tr in zip(test_results, tr_results):
        np.testing.assert_array_equal(r, r_tr)


@pytest.mark.parametrize("case", [1])
def test_copy(universe, case):
    u = universe(case)
    uc = u.copy()
    assert uc is not u
    assert u.name() == uc.name()
    assert u.verbose_name == uc.verbose_name
    assert u._comment == uc._comment
    assert len(u._cells) == len(uc._cells)
    for c, cc in zip(u, uc):
        assert cc is not c
        assert c.name() == cc.name()
        assert c.shape == cc.shape
    surfaces_idx = {s.name(): s for s in u.get_surfaces()}
    copy_surfaces_idx = {s.name(): s for s in uc.get_surfaces()}
    assert surfaces_idx == copy_surfaces_idx
    for k, v in surfaces_idx.items():
        assert copy_surfaces_idx[k] is not v
        assert copy_surfaces_idx[k] == v


@pytest.mark.slow
@pytest.mark.parametrize("tol", [0.2, None])
@pytest.mark.parametrize("case, expected", [(1, [[-10, 10], [-10, 10], [-6.5, 13.5]])])
def test_bounding_box(universe, tol, case, expected):
    u = universe(case)
    base = [0, 0, 0]
    dims = [30, 30, 30]
    gb = Box(base, dims[0], dims[1], dims[2])
    if tol is not None:
        bb = u.bounding_box(box=gb, tol=tol)
    else:
        tol = 100.0
        bb = u.bounding_box()
    for j, (low, high) in enumerate(expected):
        dimensions = 0.5 * bb.dimensions[j]
        assert bb.center[j] - dimensions <= low
        assert bb.center[j] - dimensions >= low - tol
        assert bb.center[j] + dimensions >= high
        assert bb.center[j] + dimensions <= high + tol


@pytest.mark.parametrize(
    "case, condition, inner, answer",
    [
        (1, cell_selector(1), False, [(Body, 1)]),
        (1, cell_selector([1, 3]), False, [(Body, 1), (Body, 3)]),
        (1, surface_selector(1), False, [(Surface, 1)]),
        (1, surface_selector([1, 4]), False, [(Surface, 1), (Surface, 4)]),
        (1, lambda c: [c] if c.material() else [], False, [(Body, 1), (Body, 2)]),
        (2, cell_selector(3), False, [(Body, 3)]),
        (2, cell_selector(11), False, []),
        (2, cell_selector(11), True, [(Body, 11)]),
    ],
)
def test_select(universe, case, condition, inner, answer):
    u = universe(case)
    result = u.select(condition, inner=inner)
    assert len(result) == len(answer)
    for r, (cls, name) in zip(result, answer):
        assert isinstance(r, cls)
        assert r.name() == name


@pytest.mark.parametrize(
    "case, answer",
    [(1, {0: {1, 2, 3, 4}}), (2, {0: {1, 2, 3}, 1: {10, 11, 12}, 2: {21, 22, 23, 24}})],
)
def test_get_universes(universe, case, answer):
    u = universe(case)
    universes = u.get_universes()
    assert len(universes) == len(answer.keys())
    named_u = {x.name(): x for x in universes}
    for k, v in named_u.items():
        names = {c.name() for c in v}
        assert names == answer[k]


@pytest.mark.parametrize(
    "case, rename, stat",
    [
        (1, {}, {}),
        (2, {}, {}),
        (2, {2: {"start_cell": 2}}, {"cell": {2: [0, 2], 3: [0, 2]}}),
        (2, {2: {"start_surf": 2}}, {"surf": {2: [0, 2], 3: [0, 2], 4: [0, 2]}}),
        (
            2,
            {2: {"start_cell": 2, "start_surf": 3}},
            {"cell": {2: [0, 2], 3: [0, 2]}, "surf": {3: [0, 2], 4: [0, 2]}},
        ),
        (
            2,
            {2: {"start_cell": 2}, 1: {"start_cell": 1}},
            {"cell": {1: [0, 1], 2: [0, 1, 2], 3: [0, 1, 2]}},
        ),
        (2, {1: {"name": 2}}, {"universe": {2: [1, 2]}}),
        (2, {1: {"name": 3}, 2: {"name": 3}}, {"universe": {3: [1, 2]}}),
        (2, {1: {"start_mat": 19}}, {}),
        (2, {1: {"start_mat": 20}}, {"material": {21: [1, 2]}}),
        (2, {1: {"start_mat": 21}}, {"material": {21: [1, 2], 22: [1, 2]}}),
        (
            2,
            {1: {"start_mat": 20}, 2: {"start_cell": 2}},
            {"material": {21: [1, 2]}, "cell": {2: [0, 2], 3: [0, 2]}},
        ),
        (
            2,
            {1: {"start_mat": 21}, 2: {"start_cell": 2}},
            {"material": {21: [1, 2], 22: [1, 2]}, "cell": {2: [0, 2], 3: [0, 2]}},
        ),
        (3, {}, {}),
        (4, {}, {}),
    ],
)
def test_name_clashes(
    universe,
    case: int,
    rename: Dict[int, Dict[str, int]],
    stat: TStat,
):
    rename = deepcopy(rename)
    stat = deepcopy(stat)
    u: Universe = universe(case)
    universes_index: Dict[int, Universe] = {x.name(): x for x in u.get_universes()}
    for uname, ren_dict in rename.items():
        universes_index[uname].rename(**ren_dict)
    for stat_item in stat.values():
        for kind, universe_names in stat_item.items():
            stat_item[kind] = {universes_index[uname] for uname in universe_names}
    s = u.name_clashes()
    assert s == stat


@pytest.mark.parametrize(
    "case, common_mat, stat",
    [
        (2, {Composition(atomic=[("6012", 1)], name=1)}, {}),
        (2, {Composition(atomic=[("6012", 1)], name=10)}, {}),
        (
            2,
            {Composition(atomic=[("7014", 1)], name=10)},
            {"material": {10: [1, None]}},
        ),
        (
            2,
            {
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                )
            },
            {},
        ),
        (
            2,
            {
                Composition(
                    atomic=[
                        (Element("H2", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=11,
                )
            },
            {"material": {11: [1, None]}},
        ),
        (
            2,
            {
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=21,
                )
            },
            {"material": {21: [2, None]}},
        ),
        (
            2,
            {
                Composition(atomic=[("7012", 1)], name=10),
                Composition(
                    atomic=[
                        (Element("H2", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=11,
                ),
            },
            {"material": {10: [1, None], 11: [1, None]}},
        ),
    ],
)
def test_name_clashes_with_common_materials(
    universe,
    case: int,
    common_mat: Set[Composition],
    stat: TStat,
):
    stat = deepcopy(stat)
    u = universe(case)
    universes_idx = {x.name(): x for x in u.get_universes()}
    u.set_common_materials(common_mat)
    for stat_item in stat.values():
        for kind, universes_names in stat_item.items():
            stat_item[kind] = {universes_idx[uname] if uname else None for uname in universes_names}
    s = u.name_clashes()
    for c in u._common_materials:
        print(c.mcnp_repr())
    for name, un in universes_idx.items():
        print(name)
        for c in un.get_compositions():
            print(c.mcnp_repr())
    assert s == stat


@pytest.mark.parametrize(
    "case, common_mat",
    [
        (2, {Composition(atomic=[("6012", 1)], name=1)}),
        (
            2,
            {
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                )
            },
        ),
        (
            2,
            {
                Composition(atomic=[("6012", 1)], name=1),
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                ),
            },
        ),
        (2, {Composition(atomic=[("Li", 1)], name=3)}),
    ],
)
def test_set_common_materials(universe, case, common_mat):
    u = universe(case)
    universes = u.get_universes()
    mats_before = {un: un.get_compositions() for un in universes}
    u.set_common_materials(common_mat)
    cm = {c: c for c in common_mat}
    for un in universes:
        assert mats_before[un] == un.get_compositions()
        for c in un:
            mat = c.material()
            if mat:
                comp = mat.composition
                if comp in common_mat:
                    assert comp is cm[comp]


@pytest.mark.slow
@pytest.mark.parametrize(
    "case, condition, answer_case, box",
    [
        (2, {}, 1002, Box([0, 0, 0], 20, 20, 20)),
        (2, {"cell": 1}, 1012, Box([0, 0, 0], 20, 20, 20)),
        (2, {"cell": 2}, 1022, Box([0, 0, 0], 20, 20, 20)),
        (2, {"universe": 1}, 1022, Box([0, 0, 0], 20, 20, 20)),
        (2, {"universe": 2}, 1012, Box([0, 0, 0], 20, 20, 20)),
        (
            2,
            {"predicate": lambda c: "transform" in c.options["FILL"].keys()},
            1022,
            Box([0, 0, 0], 20, 20, 20),
        ),
    ],
)
def test_apply_fill(universe, case, condition, answer_case, box):
    u = universe(case)
    ua = universe(answer_case)
    u.apply_fill(**condition)
    points = box.generate_random_points(1000000)
    test_f = u.test_points(points)
    test_a = ua.test_points(points)
    np.testing.assert_array_equal(test_f, test_a)


@pytest.mark.parametrize(
    "case, start, answer",
    [
        (
            1,
            {"name": 4},
            {
                "name": 4,
                "cell": [1, 2, 3, 4],
                "surface": [1, 2, 3, 4, 5],
                "material": [1, 2],
            },
        ),
        (
            1,
            {"start_cell": 6},
            {
                "name": 0,
                "cell": [6, 7, 8, 9],
                "surface": [1, 2, 3, 4, 5],
                "material": [1, 2],
            },
        ),
        (
            1,
            {"start_surf": 7},
            {
                "name": 0,
                "cell": [1, 2, 3, 4],
                "surface": [7, 8, 9, 10, 11],
                "material": [1, 2],
            },
        ),
        (
            1,
            {"name": 4, "start_cell": 6, "start_surf": 7},
            {
                "name": 4,
                "cell": [6, 7, 8, 9],
                "surface": [7, 8, 9, 10, 11],
                "material": [1, 2],
            },
        ),
        (
            1,
            {"name": 4, "start_mat": 5},
            {
                "name": 4,
                "cell": [1, 2, 3, 4],
                "surface": [1, 2, 3, 4, 5],
                "material": [5, 6],
            },
        ),
        (
            1,
            {"start_cell": 6, "start_mat": 6},
            {
                "name": 0,
                "cell": [6, 7, 8, 9],
                "surface": [1, 2, 3, 4, 5],
                "material": [6, 7],
            },
        ),
        (
            1,
            {"start_surf": 7, "start_mat": 4},
            {
                "name": 0,
                "cell": [1, 2, 3, 4],
                "surface": [7, 8, 9, 10, 11],
                "material": [4, 5],
            },
        ),
        (
            1,
            {"name": 4, "start_cell": 6, "start_surf": 7, "start_mat": 4},
            {
                "name": 4,
                "cell": [6, 7, 8, 9],
                "surface": [7, 8, 9, 10, 11],
                "material": [4, 5],
            },
        ),
    ],
)
def test_rename(universe, case, start, answer):
    u = universe(case)
    u.rename(**start)
    assert u.name() == answer["name"]
    cell_names = sorted(c.name() for c in u)
    surface_names = sorted(s.name() for s in u.get_surfaces())
    composition_names = sorted(m.name() for m in u.get_compositions())
    assert cell_names == answer["cell"]
    assert surface_names == answer["surface"]
    assert composition_names == answer["material"]


@pytest.mark.parametrize("case", [1, 2])
def test_alone(universe, case):
    u = universe(case)
    current_universe = u.alone()
    assert current_universe is not u
    assert 0 == current_universe.name()
    assert u.verbose_name == current_universe.verbose_name
    assert len(u._cells) == len(current_universe._cells)
    for c, cc in zip(u, current_universe):
        assert cc is not c
        assert c.name() == cc.name()
        assert c.shape == cc.shape
        assert "FILL" not in cc.options
    surfs = {s.name(): s for s in u.get_surfaces()}
    current_universe_surfs = {s.name(): s for s in current_universe.get_surfaces()}
    assert surfs == current_universe_surfs
    for k, v in surfs.items():
        assert current_universe_surfs[k] is not v
        assert current_universe_surfs[k] == v


@pytest.mark.parametrize(
    "case, common, start, answer",
    [
        (1, set(), 7, [7, 8]),
        (1, {Composition(atomic=[("C-12", 1)], name=1)}, 7, [1, 7]),
        (
            1,
            {
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                )
            },
            7,
            [2, 7],
        ),
        (
            1,
            {
                Composition(atomic=[("C-12", 1)], name=1),
                Composition(
                    atomic=[
                        (Element("H1", lib="31c"), 2 / 3),
                        (Element("O16", lib="31c"), 1 / 3),
                    ],
                    name=2,
                ),
            },
            7,
            [1, 2],
        ),
    ],
)
def test_rename_when_common_mat(universe, case, common, start, answer):
    u = Universe(universe(case), common_materials=common)
    u.rename(start_mat=start)
    composition_names = sorted(m.name() for m in u.get_compositions())
    assert composition_names == answer


@pytest.mark.slow
@pytest.mark.parametrize("verbose", [False, True])
@pytest.mark.parametrize(
    "case, complexities", [(1, {1: 1, 2: 3, 3: 5, 4: 1}), (3, {1: 1, 2: 3, 4: 5, 5: 1})]
)
def test_simplify(universe, case, complexities, verbose):
    u = universe(case)
    u.simplify(min_volume=0.1, verbose=verbose)
    assert len(u._cells) == len(complexities.keys())
    for c in u:
        assert c.shape.complexity() == complexities[c.name()]


@pytest.mark.parametrize(
    "case, box",
    [
        (1, Box([0, 0, 0], 20, 20, 20)),
        (2, Box([0, 0, 0], 20, 20, 20)),
        (3, Box([0, 0, 0], 20, 20, 20)),
    ],
)
def test_save(universe, case, box):
    u = universe(case)
    out = tempfile.NamedTemporaryFile(mode="w+b", delete=False)
    u.save(out.name)
    out.close()
    ur = from_file(out.name).universe

    points = box.generate_random_points(100000)
    universes_orig = {x.name(): x for x in u.get_universes()}
    universes_expected = {x.name(): x for x in ur.get_universes()}
    assert universes_orig.keys() == universes_expected.keys()
    for k, univ in universes_orig.items():
        test_a = univ.test_points(points)
        test_f = universes_expected[k].test_points(points)
        np.testing.assert_array_equal(test_f, test_a)


@pytest.mark.parametrize(
    "case, rename",
    [
        (2, {2: {"start_cell": 2}}),
        (2, {2: {"start_surf": 2}}),
        (2, {2: {"start_cell": 2, "start_surf": 3}}),
        (2, {2: {"start_cell": 2}, 1: {"start_cell": 1}}),
    ],
)
def test_save_exception(tmp_path, universe, case, rename):
    u = universe(case)
    universes_idx = {x.name(): x for x in u.get_universes()}
    for uname, ren_dict in rename.items():
        universes_idx[uname].rename(**ren_dict)
    with pytest.raises(NameClashError):
        u.save(tmp_path)


@pytest.mark.xfail(reason="Check this renaming")
@pytest.mark.parametrize(
    "case, rename",
    [
        (2, {1: {"name": 2}}),
        (2, {1: {"name": 3}, 2: {"name": 3}}),
    ],
)
def test_save_exception2(tmp_path, universe, case, rename):
    u = universe(case)
    universes_idx = {x.name(): x for x in u.get_universes()}
    for uname, ren_dict in rename.items():
        universes_idx[uname].rename(**ren_dict)
    with pytest.raises(NameClashError):
        u.save(tmp_path)


@pytest.mark.parametrize(
    "case, expected",
    [
        (
            """\
                0
                1 0 -1
                2 0  1

                1 1 so 1

                TR1 1 1 1
            """,
            [1],
        ),
        (
            """\
            0
            1 0 -1
            2 0  1

            1 1 so 1

            TR1 1 1 1
            """,
            [1],
        ),
        (
            """\
            0
            1 0 -1 -2
            2 0  1

            1 1 so 2
            2 2 so 3

            TR1 1 1 1
            TR2 -1 -1 -1
            """,
            [1, 2],
        ),
        (
            """\
            0
            1 0 -1 -2
            2 0  1 fill=1 (3) $ fill with named transformation
            3 0  3 u=1

            1 1 so 2
            2 2 so 3
            3 so 4

            TR1 1 1 1
            TR2 -1 -1 -1
            TR3  0 1 0
            """,
            [1, 2, 3],
        ),
    ],
)
def test_collect_transformations(case: str, expected: List[int]) -> None:
    case = textwrap.dedent(case)
    u = from_text(case).universe
    actual = sorted(map(int, map(Card.name, collect_transformations(u))))
    assert actual == expected



from typing import Iterable, Optional

from mckit.material import Composition, Material
from mckit.utils.indexes import Index, NumberedItemNotFoundError


class DummyMaterial(Material):
    def __init__(
        self,
        name: int,
        *,
        density: Optional[float] = None,
        concentration: Optional[float] = None,
    ) -> None:
        assert (density is None) ^ (concentration is None), "Specify only one of the parameters"
        if density is None:
            # noinspection PyTypeChecker
            super().__init__(
                composition=DummyComposition(name), concentration=concentration * 1.0e24
            )
        else:
            super().__init__(composition=DummyComposition(name), density=density)


# noinspection PyTypeChecker
class DummyComposition(Composition):
    """To substitute composition when it's not found."""

    def __init__(self, name: int):
        super().__init__(name=name, weight=[(1001, 1.0)], comment="dummy")


def raise_on_absent_composition_strategy(name: int) -> Optional[DummyComposition]:
    raise CompositionNotFoundError(name)


def dummy_on_absent_composition_strategy(name: int) -> Optional[DummyComposition]:
    return DummyComposition(name)


class CompositionStrictIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(raise_on_absent_composition_strategy, **kwargs)

    @classmethod
    def from_iterable(cls, items: Iterable[Composition]) -> "CompositionStrictIndex":
        index = cls()
        index.update((c.name(), c) for c in items)
        return index


class CompositionDummyIndex(Index):
    def __init__(self, **kwargs):
        super().__init__(dummy_on_absent_composition_strategy, **kwargs)


class CompositionNotFoundError(NumberedItemNotFoundError):
    kind = "Composition"



import mckit.parser.common.transformation_index as ti
import mckit.parser.surface_parser as srp
import pytest

from mckit.surface import create_surface
from mckit.utils.indexes import IgnoringIndex


@pytest.mark.parametrize(
    "text,expected",
    [
        ("1 PX 0", [("INTEGER", 1), ("SURFACE_TYPE", "PX"), ("INTEGER", 0)]),
        (
            "+2 py 2",
            [("MODIFIER", "+"), ("INTEGER", 2), ("SURFACE_TYPE", "PY"), ("INTEGER", 2)],
        ),
        (
            """
    158214   RPP  +558.364973080895990 +902.003588792269060  &
              -428.589771011879980 -105.426621291914000  &
              +438.709115982056010 +662.547063827514990
                """,
            [
                ("INTEGER", 158214),
                ("SURFACE_TYPE", "RPP"),
                ("FLOAT", 558.364973080895990),
                ("FLOAT", 902.003588792269060),
                ("FLOAT", -428.589771011879980),
                ("FLOAT", -105.426621291914000),
                ("FLOAT", 438.709115982056010),
                ("FLOAT", 662.547063827514990),
            ],
        ),
    ],
)
def test_surface_lexer(text, expected):
    lexer = srp.Lexer()
    actual = [(f.type, f.value) for f in lexer.tokenize(text)]
    assert actual == expected


# noinspection PyTypeChecker
@pytest.mark.parametrize(
    "text,expected",
    [
        ("1 PX 0", create_surface("PX", 0.0, name=1)),
        (
            "*1 P 1.5 1.4 1.3 1.2",
            create_surface("P", 1.5, 1.4, 1.3, 1.2, name=1, modifier="*"),
        ),
        (
            """
    158214   RPP  +558.364973080895990 +902.003588792269060  &
              -428.589771011879980 -105.426621291914000  &
              +438.709115982056010 +662.547063827514990
         """,
            create_surface(
                "RPP",
                558.364973080895990,
                902.003588792269060,
                -428.589771011879980,
                -105.426621291914000,
                438.709115982056010,
                662.547063827514990,
                name=158214,
            ),
        ),
    ],
)
def test_good_path(text, expected):
    actual = srp.parse(text)
    assert actual.mcnp_words() == expected.mcnp_words()


# noinspection PyTypeChecker
@pytest.mark.parametrize("text, expected", [("1 2 PX 0", create_surface("PX", 0.0))])
def test_absent_transformation_with_ignore_strategy(text, expected):
    actual = srp.parse(text, transformations=IgnoringIndex())
    assert actual == expected
    assert "transform" not in actual.options


@pytest.mark.parametrize("text,msg_contains", [("1 2 PX 0", 2)])
def test_absent_surface_with_raise_strategy(text, msg_contains):
    with pytest.raises(KeyError, match=f"Transformation #{msg_contains} is not found"):
        srp.parse(text, transformations=ti.TransformationStrictIndex())


# noinspection PyTypeChecker
@pytest.mark.parametrize(
    "text,expected",
    [
        ("1 2 PX 0.1", create_surface("PX", 0.1, transform=ti.DummyTransformation(2))),
        (
            "1 1 SX 4 +5.0",
            create_surface("SX", 4, 5.0, name=1, transform=ti.DummyTransformation(1)),
        ),
    ],
)
def test_absent_surface_with_dummy_strategy(text, expected):
    actual = srp.parse(text, transformations=ti.TransformationDummyIndex())
    assert actual == expected


if __name__ == "__main__":
    pytest.main()



import re

import numpy as np
import ply.lex as lex
import ply.yacc as yacc

from ..fmesh import FMesh

literals = ["+", "-", ":", "/"]

KEYWORDS = [
    "MCNP",
    "VERSION",
    "LD",
    "PROBID",
    "NEUTRON",
    "PHOTON",
    "ELECTRON",
    "RESULT",
    "RESULTS",
    "ERROR",
    "ERRORS",
    "CYLINDER",
    "ORIGIN",
    "AXIS",
    "TALLY",
    "MPI",
    "X",
    "Y",
    "Z",
    "R",
    "THETA",
    "ENERGY",
    "TIME",
    "TH",
    "TOTAL",
]

BIN_REC_ORDER = {"ENERGY": 0, "X": 1, "Y": 2, "Z": 3, "TIME": 0}
BIN_CYL_ORDER = {"ENERGY": 0, "R": 1, "Z": 2, "THETA": 3, "TIME": 0}

# List of token names
tokens = [
    "separator",
    "newline",
    "int_number",
    "flt_number",
    "minus",
    "title",
    "stamp",
    "MCNP",
    "VERSION",
    "LD",
    "PROBID",
    "NEUTRON",
    "PHOTON",
    "ELECTRON",
    "RESULT",
    "ERROR",
    "CYLINDER",
    "ORIGIN",
    "AXIS",
    "TALLY",
    "MPI",
    "X",
    "Y",
    "Z",
    "R",
    "THETA",
    "ENERGY",
    "TIME",
    "TOTAL",
]

# precedence = (
#    ('left', 'newline'),
# )

states = (("title", "exclusive"), ("norm", "exclusive"), ("tally", "exclusive"))

BLANK_LINE = r"^\n"
STAMP = r"[\d/:]+"
NEWLINE = r"\n"
FRACTION = r"\."
EXPONENT = r"(E[-+]?\d+)"
INT_NUMBER = r"(\d+)"
MINUS = r"-(?=\d)"
FLT_NUMBER = (
    r"("
    + INT_NUMBER
    + r"?"
    + FRACTION
    + INT_NUMBER
    + EXPONENT
    + r"?|"
    + INT_NUMBER
    + FRACTION
    + r"?"
    + EXPONENT
    + r"|"
    + INT_NUMBER
    + FRACTION
    + r")(?=[ \n-+])"
)
KEYWORD = r"[A-Z]+(/[A-Z]+)?"

SKIP = r"[=,.() ]"

t_ANY_ignore = SKIP


def t_title_title(t):
    r""".+"""
    return t


@lex.TOKEN(MINUS)
def t_norm_tally_minus(t):
    return t


# noinspection PyPep8Naming
@lex.TOKEN(BLANK_LINE)
def t_ANY_separator(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    return t


@lex.TOKEN(NEWLINE)
def t_newline(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.begin("title")
    return t


@lex.TOKEN(NEWLINE)
def t_title_newline(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.begin("norm")
    return t


@lex.TOKEN(NEWLINE)
def t_norm_newline(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    t.lexer.begin("tally")
    return t


@lex.TOKEN(NEWLINE)
def t_tally_newline(t):
    t.lexer.lineno += 1
    t.lexer.last_pos = t.lexer.lexpos
    return t


@lex.TOKEN(STAMP)
def t_stamp(t):
    return t


@lex.TOKEN(KEYWORD)
def t_keyword(t):
    value = t.value.upper()
    if value == "MCNP":
        t.lexer.section_index = 0
        meshtal_lexer.lineno = 1
        t.lineno = 1
        t.lexer.last_pos = 1
    if value in KEYWORDS:
        t.type = value
        t.value = value
        return t


@lex.TOKEN(KEYWORD)
def t_norm_tally_keyword(t):
    value = t.value.upper()
    if value == "TIMES":
        value = "TIME"
    if value in KEYWORDS:
        if value == "RESULTS":
            value = "RESULT"
        elif value == "ERRORS":
            value = "ERROR"
        elif value == "TH":
            value = "THETA"
        t.type = value
        t.value = value
        return t


# noinspection PyPep8Naming
def t_ANY_error(t):
    column = t.lexer.lexpos - t.lexer.last_pos + 1
    msg = r"Illegal character '{0}' at line {1} column {2}".format(
        t.value[0], t.lexer.lineno, column
    )
    raise ValueError(msg, t.value[0], t.lexer.lineno, column)


@lex.TOKEN(FLT_NUMBER)
def t_norm_tally_flt_number(t):
    t.value = float(t.value)
    return t


@lex.TOKEN(INT_NUMBER)
def t_norm_tally_int_number(t):
    t.value = int(t.value)
    return t


meshtal_lexer = lex.lex(reflags=re.MULTILINE + re.IGNORECASE + re.VERBOSE)


def p_meshtal(p):
    """meshtal : header newline title newline float newline separator tallies"""
    p[0] = {
        "date": p[1]["PROBID"],
        "title": p[3].lstrip().rstrip(),
        "histories": p[5],
        "tallies": p[8],
    }


def p_header(p):
    """header : MCNP VERSION stamp LD stamp PROBID stamp stamp
    | MCNP VERSION stamp MPI LD stamp PROBID stamp stamp
    """
    l = len(p)
    p[0] = {"PROBID": p[l - 2] + p[l - 1], "VERSION": p[3]}


def p_tallies(p):
    """tallies : tallies tally
    | tallies tally separator
    | tally separator
    | tally
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        if isinstance(p[1], list):
            p[1].append(p[2])
            p[0] = p[1]
        else:
            p[0] = [p[1]]


def p_tally(p):
    """tally : tally_header separator boundaries separator data"""
    tally = p[1]
    boundaries = p[3]
    tally["geom"] = "XYZ"
    if "ORIGIN" in boundaries.keys():
        tally["origin"] = boundaries.pop("ORIGIN")
        tally["geom"] = "CYL"
    if "AXIS" in boundaries.keys():
        tally["axis"] = boundaries.pop("AXIS")
    tally["bins"] = {k: np.array(v) for k, v in boundaries.items()}
    data = p[5]
    od = BIN_CYL_ORDER if "origin" in tally.keys() else BIN_REC_ORDER
    if "result" in data.keys():
        src_perm = [od[let] for let in data["order"]]
        tally["result"] = np.moveaxis(np.array(data["result"]), (0, 1, 2, 3), src_perm)
        tally["error"] = np.moveaxis(np.array(data["error"]), (0, 1, 2, 3), src_perm)
    else:
        header = data["header"]
        data = np.array(data["data"])
        shape = [0, 0, 0, 0]
        for k in tally["bins"].keys():
            v = od[k]
            shape[v] = boundaries[k].size
            if k != "TIME":
                shape[v] -= 1
        if "ENERGY" in boundaries.keys():
            boundaries["ENERGY"] = 0.5 * (boundaries["ENERGY"][1:] + boundaries["ENERGY"][:-1])
        result = np.empty(shape)
        error = np.empty(shape)
        indices = np.empty((data.shape[0], 4), dtype=int)
        for k in boundaries.keys():
            if k in header:
                indices[:, od[k]] = np.searchsorted(boundaries[k], data[:, header.index(k)]) - 1
            else:
                indices[:, od[k]] = np.zeros(data.shape[0])
        res_ind = header.index("RESULT")
        err_ind = header.index("ERROR")
        for i in range(indices.shape[0]):
            result[tuple(indices[i, :])] = data[i, res_ind]
            error[tuple(indices[i, :])] = data[i, err_ind]
        tally["result"] = result
        tally["error"] = error
    p[0] = tally


def p_tallY_header(p):
    """tally_header : TALLY integer newline particle TALLY newline
    | TALLY integer newline particle TALLY newline ENERGY newline
    | TALLY integer newline particle TALLY newline TALLY newline
    | TALLY integer newline particle TALLY newline TALLY newline ENERGY newline
    """
    p[0] = {"name": p[2], "particle": p[4]}


def p_particle(p):
    """particle : NEUTRON
    | PHOTON
    | ELECTRON
    """
    p[0] = p[1]


def p_boundaries(p):
    """boundaries : TALLY ':' newline CYLINDER ORIGIN vector AXIS vector newline bins
    | TALLY ':' newline bins
    """
    boundaries = p[len(p) - 1]
    if len(p) == 11:
        boundaries["ORIGIN"] = p[6]
        boundaries["AXIS"] = p[8]
    for k, v in boundaries.items():
        boundaries[k] = np.array(v)
    p[0] = boundaries


def p_bins(p):
    """bins : direction newline direction newline direction newline energies newline"""
    bins = {}
    for name, data in [p[1], p[3], p[5], p[7]]:
        bins[name] = data
    p[0] = bins


def p_dir_spec(p):
    """dir_spec : X
    | Y
    | Z
    | R
    | THETA
    """
    p[0] = p[1]


def p_direction(p):
    """direction : dir_spec ':' vector"""
    p[0] = p[1], p[3]


def p_energies(p):
    """energies : ENERGY ':' vector
    | TIME ':' vector
    """
    p[0] = p[1], p[3]


def p_float(p):
    """float : '+' flt_number
    | minus flt_number
    | flt_number
    | integer
    """
    if p[1] == "-":
        p[0] = -p[2]
    elif p[1] == "+":
        p[0] = p[2]
    else:
        p[0] = p[1]


def p_integer(p):
    """integer : '+' int_number
    | minus int_number
    | int_number
    """
    if p[1] == "-":
        p[0] = -p[2]
    elif p[1] == "+":
        p[0] = p[2]
    else:
        p[0] = p[1]


def p_vector(p):
    """vector : vector float
    | float
    """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[1].append(p[2])
        p[0] = p[1]


def p_matrix(p):
    """matrix : matrix vector newline
    | vector newline
    """
    if len(p) == 3:
        p[0] = [p[1]]
    else:
        p[1].append(p[2])
        p[0] = p[1]


def p_total_matrix(p):
    """total_matrix : total_matrix TOTAL vector newline
    | TOTAL vector newline"""
    if len(p) == 4:
        p[0] = [p[2]]
    else:
        p[1].append(p[4])
        p[0] = p[1]


def p_data(p):
    """data : matrix_data
    | column_data separator
    | column_data
    """
    p[0] = p[1]


def p_column_data(p):
    """column_data : column_header matrix total_matrix
    | column_header matrix
    """
    p[0] = {"header": p[1], "data": p[2]}
    if len(p) == 4:
        p[0]["total"] = p[3]


def p_column_header(p):
    """column_header : ENERGY dir_spec dir_spec dir_spec RESULT ERROR newline
    | dir_spec dir_spec dir_spec RESULT ERROR newline
    """
    n = len(p)
    p[0] = p[1 : n - 1]


def p_matrix_data(p):
    """matrix_data : energy_bins total_energy_bin
    | energy_bins
    """
    order, result, error = p[1]
    p[0] = {"result": result, "error": error, "order": order}


def p_energy_bins(p):
    """energy_bins : energy_bins energy_bin separator
    | energy_bin separator
    """
    if len(p) == 3:
        order, result, error = p[1]
        p[0] = order, [result], [error]
    else:
        order, result, error = p[2]
        p[1][1].append(result)
        p[1][2].append(error)
        p[0] = order, p[1][1], p[1][2]


def p_energy_bin(p):
    """energy_bin : ENERGY ':' float '-' float newline separator spatial_bins
    | TIME ':' float newline separator spatial_bins
    """
    l = len(p) - 1
    order = [p[1]] + p[l][0]
    p[0] = order, p[l][1], p[l][2]


def p_total_energy_bin(p):
    """total_energy_bin : TOTAL ENERGY newline separator spatial_bins separator"""
    order = [p[1]] + p[5][0]
    p[0] = order, p[5][1], p[5][2]


def p_spatial_bins(p):
    """spatial_bins : spatial_bins spatial_bin separator
    | spatial_bin separator
    """
    if len(p) == 3:
        order, result, error = p[1]
        p[0] = order, [result], [error]
    else:
        order, result, error = p[2]
        p[1][1].append(result)
        p[1][2].append(error)
        p[0] = order, p[1][1], p[1][2]


def p_spatial_bin(p):
    """spatial_bin : dir_spec ':' float '-' float newline separator TALLY RESULT ':' dir_spec dir_spec newline matrix separator ERROR newline matrix separator"""
    order = [p[1], p[12], p[11]]
    results = [line[1:] for line in p[14][1:]]
    errors = [line[1:] for line in p[18][1:]]
    p[0] = order, results, errors


def p_error(p):
    if p:
        column = p.lexer.lexpos - p.lexer.last_pos + 1
        print(
            "Syntax error at token {0} {3}, line {1}, column {2}".format(
                p.type, p.lexer.lineno, column, p.value
            )
        )
    else:
        print("Syntax error at EOF")


meshtal_parser = yacc.yacc(tabmodule="meshtal_tab", debug=True)


_BIN_NAMES = {
    "ENERGY": "ebins",
    "X": "xbins",
    "Y": "ybins",
    "Z": "zbins",
    "R": "rbins",
    "THETA": "tbins",
    "TIME": "dtbins",
}


def read_meshtal(filename):
    """Reads MCNP meshtal file.

    Parameters
    ----------
    filename : str
        File that contains MCNP meshtally data.

    Returns
    -------
    tallies : dict
        Dictionary of mesh tallies contained in the file. It is
        tally_name -> Fmesh pairs.
    """
    with open(filename) as f:
        text = f.read() + "\n"
    meshtal_lexer.begin("INITIAL")
    meshtal_data = meshtal_parser.parse(text, lexer=meshtal_lexer)
    histories = meshtal_data["histories"]
    tallies = {}
    for t in meshtal_data["tallies"]:
        name = t["name"]
        particle = t["particle"]
        data = t["result"]
        error = t["error"]
        kwdata = {}
        for k, v in t["bins"].items():
            kwdata[_BIN_NAMES[k]] = v
        tallies[name] = FMesh(name, particle, data, error, histories=histories, **kwdata)
    return tallies



import re

import pytest

from mckit import __version__

VERSION_PATTERN = re.compile(r"\d+\.\d+\.\d+.*")


def test_version():
    assert VERSION_PATTERN.match(__version__)
    assert __version__ != "0.0.0", "meta information is not correct"


if __name__ == "__main__":
    pytest.main()



"""Code for transformations."""
from __future__ import annotations

from typing import Any, Optional, Tuple

import numpy as np

# noinspection PyUnresolvedReferences,PyPackageRequirements
from mckit.geometry import ORIGIN
from numpy.typing import ArrayLike

from .card import Card
from .utils import compute_hash
from .utils.tolerance import DEFAULT_TOLERANCE_ESTIMATOR, EstimatorType, MaybeClose

__all__ = ["Transformation", "IDENTITY_ROTATION"]

IDENTITY_ROTATION = np.eye(3)

ANGLE_TOLERANCE = 0.001
COS_TH = np.sin(ANGLE_TOLERANCE)
ZERO_COS_TOLERANCE = 2.0e-16


class Transformation(Card, MaybeClose):
    """Geometry transformation object.

    Parameters
    ----------
    translation : array_like[float]
        Translation vector of the transformation. Default: zero vector - no
        translation.
    rotation : array_like[float]
        Represents rotation matrix of the transformation. It can be either
        matrix itself with shape 3x3 or a list of transformation parameters.
        In the last case the parameters are treated as follows: 1-3 - ex', 4-6 -
        ey', 7-9 - ez', where ex', ey' and ez' - basis vectors of the local
        coordinate system. Basis vectors should be pairwise orthogonal. If there
        is a small non-orthogonality - less than 0.001 radian - vectors are
        corrected. Default: identity matrix - non rotation.
    indegrees : bool
        How rotation parameters should be treated. If True - rotation matrix
        parameters are given in degrees. Otherwise rotation parameters are basis
        vectors of local coordinate system. Default: False.
    inverted : bool
        How translation vector should be interpreted. If True - it is the origin
        of local coordinate system defined in the global one. Otherwise - the
        origin of global coordinate system defined in the local one.
        Default: False
    options : dict
        Other options, like name, comment, etc.

    Methods
    -------
    apply2gq(m, v, k)
        Gets parameters of generic quadratic surface in the main coordinate
        system.
    apply2plane(v, k)
        Gets parameters of plane in the main coordinate system.
    apply2point(p)
        Gets coordinates of point p in the global coordinate system.
    apply2vector(v)
        Gets coordinates of vector v in the global coordinate system.
    apply2transform(tr)
        Gets resulting transformation.
    reverse()
        Reverses this transformation, and returns the result.
    """

    def __init__(
        self,
        translation=ORIGIN,
        rotation=None,
        indegrees=False,
        inverted=False,
        **options: dict[str, Any],
    ):

        Card.__init__(self, **options)

        if translation is not ORIGIN:
            translation = np.asarray(translation, dtype=float)
            if translation.shape != (3,):
                raise ValueError(f"Transaction #{self.name()}: wrong length of translation vector.")

        if rotation is None:
            u = IDENTITY_ROTATION
        else:
            u = np.asarray(rotation, dtype=float)
            if indegrees:
                u = np.cos(np.multiply(u, np.pi / 180.0))
            u = self._setup_rotation_matrix(u)
        self._u = u
        if inverted:
            if u is IDENTITY_ROTATION:
                self._t = -translation
            else:
                self._t = -np.dot(u, translation)
        else:
            self._t = translation.copy()
        self._hash = compute_hash(self._t, self._u)

    def mcnp_words(self, pretty: bool = False) -> list[str]:
        """Prepare words to output this Transformation with MCNP label."""
        name = self.name()
        if name is None:
            name = 0
        words = ["*", f"TR{name}"]
        words.extend(self.get_words(pretty))
        return words

    def get_words(self, _pretty: bool = False) -> list[str]:
        """Prepare words to output this Transformation content."""
        words = []
        for v in self._t:
            words.append(" ")
            words.append(f"{v:.10g}")  # TODO dvp: check if precision 13 is necessary
        if self._u is not IDENTITY_ROTATION:
            for v in self._u.transpose().ravel():
                words.append(" ")
                _value = np.arccos(v) * 180 / np.pi
                words.append(f"{_value:.10g}")
        return words

    def apply2gq(
        self, m1: ArrayLike, v1: ArrayLike, k1: float
    ) -> Tuple[ArrayLike, ArrayLike, float]:
        """Gets parameters of generic quadratic surface in the main CS.

        Args:
            m1 : array_like[float]
                A 3x3 matrix which defines coefficients of quadratic terms of
                generic quadratic surface equation in the auxiliary coordinate
                system.
            v1 : array_like[float]
                A vector of size 3, which defines coefficients of linear terms of
                generic quadratic equation in the auxiliary coordinate system.
            k1 : float
                Free term of generic quadratic equation in the auxiliary coordinate
                system.

        Returns:
            m, v, k
            m : numpy.ndarray
                A 3x3 matrix which defines quadratic coefficients of GQ surface
                equation in the main coordinate system.
            v : numpy.ndarray
                A vector of size 3 which defines linear coefficients of GQ surface
                equation in the main coordinate system.
            k : float
                Free term of GQ surface equation in the main coordinate system.
        """
        m = np.dot(np.dot(self._u, m1), np.transpose(self._u))
        v = np.dot(self._u, v1) - 2 * np.dot(m, self._t)
        k = k1 - np.dot(v, self._t) - np.dot(self._t, np.dot(m, self._t))
        return m, v, k

    def apply2plane(self, v1: ArrayLike, k1: float) -> Tuple[ArrayLike, float]:
        """Gets parameters of plane surface in the main coordinate system.

        Args:
            v1: A vector of size 3 which defines vector, normal to the plane in
                the auxiliary coordinate system.
            k1: Free term of plane equation in the auxiliary coordinate
                system.

        Returns:
            v, k
            - v: A vector of size 3 which defines vector, normal to the plane surface
                 in the main coordinate system.
            - k: Free term of plane surface equation in the main coordinate system.
        """
        v = np.dot(self._u, v1)
        k = k1 - np.dot(v, self._t)
        return v, k

    def apply2point(self, p1: ArrayLike) -> ArrayLike:
        """Gets coordinates of point p1 in the main coordinate system.

        Args:
            p1: Coordinates of the point(s) in the auxiliary coordinate system.
                It has shape (3,) if there is the only point or (N, 3) - if there
                are N points.

        Returns:
            Coordinates of the point(s) in the main coordinate system.
        """
        # Matrix U is transposed to change U p1 -> p1 U^T - to preserve shape
        # of p1 and p.
        return np.dot(p1, np.transpose(self._u)) + self._t

    def apply2vector(self, v1: ArrayLike) -> ArrayLike:
        """Gets coordinates of vector v1 in the main coordinate system.

        Args:
            v1: Coordinates of the vector(s) in the auxiliary coordinate system.

        Returns:
            Coordinates of the vector(s) in the main coordinate system.
        """
        # In contrast with apply2point - no translation is needed.
        return np.dot(v1, np.transpose(self._u))

    def apply2transform(self, tr: "Transformation") -> "Transformation":
        """Gets new transformation.

        Suppose there are three coordinate systems r, r1, r2. Transformation
        tr: r2 -> r1; and this transformation: r1 -> r. Thus the resulting
        transformation: r2 -> r. In other words the result is a sequence of
        two transformations: `tr` and this. The `tr` is applied first.

        Args:
            tr: Transformation to be modified.

        Returns:
            New transformation - the result.
        """
        rot = np.dot(self._u, tr._u)
        trans = self.apply2point(tr._t)
        return Transformation(translation=trans, rotation=rot)

    def reverse(self) -> "Transformation":
        """Reverses this transformation.

        Gets new transformation which is complement to this one.

        Returns:
            Reversed version of this transformation.
        """
        u1 = np.transpose(self._u)
        t1 = -np.dot(u1, self._t)
        return Transformation(translation=t1, rotation=u1)

    def is_close_to(
        self, other: object, estimator: EstimatorType = DEFAULT_TOLERANCE_ESTIMATOR
    ) -> bool:
        """Check if this transformation is almost equal to `other` one."""
        if self is other:
            return True
        if not isinstance(other, Transformation):
            return False
        return estimator((self._t, self._u), (other._t, other._u))

    def _setup_rotation_matrix(self, u: Optional[ArrayLike]) -> ArrayLike:
        zero_cosines_idx = np.abs(u) < ZERO_COS_TOLERANCE
        u[zero_cosines_idx] = 0.0
        # TODO: Implement creation from reduced rotation parameter set.
        if u.shape == (9,):
            u = u.reshape((3, 3), order="F")
        if u.shape != (3, 3):
            raise ValueError(
                f'Transaction{"" if self.is_anonymous else " #" + str(self.name())}: \
                  wrong number of rotation parameters: {u}.'
            )

        if np.array_equal(u, IDENTITY_ROTATION):
            return IDENTITY_ROTATION

        # normalize auxiliary CS basis and orthogonalize it.
        u, r = np.linalg.qr(u)
        # QR decomposition returns orthogonal matrix u - which is corrected
        # rotation matrix, and upper triangular matrix r. On the main
        # diagonal r contains lengths of corresponding (negative if the
        # corrected vector is directed opposite to the initial one) input
        # basis vectors. Other elements are cosines of angles between
        # different basis vectors.

        # cos(pi/2 - ANGLE_TOLERANCE) = sin(ANGLE_TOLERANCE) - maximum
        # value of cosine of angle between two basis vectors.
        if abs(r[0, 1]) > COS_TH or abs(r[0, 2]) > COS_TH or abs(r[1, 2]) > COS_TH:
            raise ValueError(
                f"Transaction #{self.name()}: non-orthogonality is greater than 0.001 rad."
            )
        # To preserve directions of corrected basis vectors.
        for i in range(3):
            u[:, i] = u[:, i] * np.sign(r[i, i])

        return u

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other) -> bool:
        return self is other or (
            isinstance(other, Transformation)
            and np.array_equal(self._u, other._u)
            and np.array_equal(self._t, other._t)
        )

    def __repr__(self):
        return f"Transformation(translation={self._t}, rotation={self._u})"

    def __getitem__(self, key):
        return self.options.get(key, None)

    def __setitem__(self, key, value):
        self.options[key] = value



# Set output for parser debugging before importing Parser classes.
# from sly import Parser as SlyParser
# SlyParser.debugfile = "sly-debug.out"
from typing import List, Optional

import mckit.parser.cell_parser as clp
import pytest

from mckit.body import Body
from mckit.material import Material
from mckit.parser.common import (
    CellDummyIndex,
    CellStrictIndex,
    CompositionStrictIndex,
    DummyComposition,
    DummyMaterial,
    DummySurface,
    SurfaceDummyIndex,
    SurfaceStrictIndex,
)
from mckit.transformation import Transformation
from mckit.utils.indexes import Index


def create_cell(
    cell_no: int,
    geometry: List,
    material: Optional[Material] = None,
    transformation: Optional[Transformation] = None,
    **options,
):
    if not options:
        options = dict()
    options["name"] = cell_no

    def convert_integers_to_surfaces(_geometry):
        for i, e in enumerate(_geometry):
            if isinstance(e, int):
                _geometry[i] = DummySurface(e)
            elif isinstance(e, list):
                convert_integers_to_surfaces(e)

    convert_integers_to_surfaces(geometry)

    if material:
        options["MAT"] = material
    if transformation:
        options["TRCL"] = transformation

    return Body(geometry, **options)


@pytest.mark.parametrize(
    "text,expected_types,expected_values",
    [
        ("1 0", ["INTEGER", "ZERO"], [1, 0]),
        ("1 1 -0.083", ["INTEGER", "INTEGER", "FLOAT"], [1, 1, -0.083]),
    ],
)
def test_cell_lexer(text, expected_types, expected_values):
    lexer = clp.Lexer()
    tokens = list(lexer.tokenize(text))
    actual_types = list(f.type for f in tokens)
    actual_values = list(f.value for f in tokens)
    assert actual_types == expected_types
    assert actual_values == expected_values


@pytest.mark.parametrize(
    "text,expected,surfaces",
    [
        ("1 0 1", create_cell(1, [1]), [1]),
        ("1 0 -1", create_cell(1, [1, "C"]), [1]),
        ("1 0 1 2", create_cell(1, [1, 2, "I"]), [1, 2]),
        ("1 0 1 -2", create_cell(1, [1, 2, "C", "I"]), [1, 2]),
        ("1 0 1  2 -3", create_cell(1, [1, 2, "I", 3, "C", "I"]), [1, 2, 3]),
        ("1 0 (1  2) : 3", create_cell(1, [1, 2, "I", 3, "U"]), [1, 2, 3]),
        (
            "1 0 (1  2) : (3 4)",
            create_cell(1, [1, 2, "I", 3, 4, "I", "U"]),
            [1, 2, 3, 4],
        ),
    ],
)
def test_parser_geometry(text, expected, surfaces):
    surfaces_index = create_dummy_surface_index(surfaces)
    actual = clp.parse(text, surfaces=surfaces_index)
    expected.options["original"] = text
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected,surfaces,compositions",
    [
        (
            "1 1 -1.0 1",
            create_cell(1, [1], material=DummyMaterial(1, density=1.0)),
            [1],
            [1],
        ),
        (
            "1 1 .08 1",
            create_cell(1, [1], material=DummyMaterial(1, concentration=0.08)),
            [1],
            [1],
        ),
    ],
)
def test_parser_with_materials(text, expected, surfaces, compositions):
    surfaces_index = create_dummy_surface_index(surfaces)
    composition_index = create_dummy_composition_index(compositions)
    actual = clp.parse(text, surfaces=surfaces_index, compositions=composition_index)
    expected.options["original"] = text
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected,surfaces",
    [
        ("1 0 1 IMP:n=1.0", create_cell(1, [1], **{"IMPN": 1.0}), [1]),
        ("1 0 1 vol 1.0", create_cell(1, [1], **{"VOL": 1.0}), [1]),
        ("1 0 1 U=200 PMT=0", create_cell(1, [1], **{"U": 200, "PMT": 0}), [1]),
    ],
)
def test_parser_with_attributes(text, expected, surfaces):
    surfaces_index = create_dummy_surface_index(surfaces)
    actual = clp.parse(text, surfaces=surfaces_index)
    expected.options["original"] = text
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected,surfaces,cells",
    [
        (
            "2 like 1 but imp:p=2.0",
            create_cell(2, [1], **{"IMPN": 1.0, "IMPP": 2.0}),
            [1],
            [create_cell(1, [1], **{"IMPN": 1.0})],
        )
    ],
)
def test_parser_with_like_spec(text, expected, surfaces, cells):
    surfaces_index = create_dummy_surface_index(surfaces)
    cells_index = CellStrictIndex.from_iterable(cells)
    actual = clp.parse(text, cells=cells_index, surfaces=surfaces_index)
    expected.options["original"] = text
    assert actual == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        (
            """93    0   (
            16 24
          )
C
          (
            -565:+563
          )
        """,
            93,
        )
    ],
)
def test_found_failures(text, expected):
    surfaces_index = SurfaceDummyIndex()
    cells_index = CellDummyIndex()
    actual = clp.parse(text, cells=cells_index, surfaces=surfaces_index)
    assert actual is not None
    assert actual.name() == expected


def create_dummy_surface_index(surfaces: List[int]) -> Index:
    return SurfaceStrictIndex.from_iterable(map(DummySurface, surfaces))


def create_dummy_composition_index(compositions: List[int]) -> Index:
    return CompositionStrictIndex.from_iterable(map(DummyComposition, compositions))


if __name__ == "__main__":
    pytest.main()



from pathlib import Path

import numpy as np
import pytest

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

from mckit.cli.commands.decompose import get_default_output_directory
from mckit.cli.runner import mckit
from mckit.parser import from_file
from mckit.utils.resource import filename_resolver
from numpy.testing import assert_array_equal

data_filename_resolver = filename_resolver("tests")


@pytest.mark.parametrize(
    "path, expected_cells",
    [
        ("cli/data/simple_cubes.mcnp", 3),
        ("cli/data/simple_cubes.universes/envelopes.i", 3),
        ("cli/data/simple_cubes.universes/u1.i", 2),
        ("cli/data/simple_cubes.universes/u2.i", 2),
    ],
)
def test_input_files_reading(path, expected_cells):
    universe = from_file(data_filename_resolver(path)).universe
    assert len(universe) == expected_cells, f"Failed to read from file {path}"


def test_help_decompose(runner):
    result = runner.invoke(mckit, args=["decompose", "--help"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "Usage: mckit decompose" in result.output


def test_when_there_is_no_args(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["decompose"], catch_exceptions=False)
        assert result.exit_code != 0, "Should fail when no arguments provided"
        assert "Usage:" in result.output


def test_not_existing_mcnp_file(runner):
    result = runner.invoke(mckit, args=["decompose", "not-existing.imcnp"], catch_exceptions=False)
    assert result.exit_code > 0
    assert "Path 'not-existing.imcnp' does not exist" in result.output


@pytest.mark.parametrize("source, expected", [("parser_test_data/parser1.txt", "envelopes.i")])
def test_when_there_are_no_universes(runner, source, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem():
        result = runner.invoke(
            mckit, args=["decompose", "-o", "universes", source], catch_exceptions=False
        )
        assert result.exit_code == 0, "Should success without universes"
        assert Path(
            "universes/envelopes.i"
        ).exists(), "Should store the only envelopes.i file in the default directory 'universes'"


@pytest.mark.parametrize(
    "source,expected", [("cli/data/simple_cubes.mcnp", "envelopes.i u1.i u2.i".split())]
)
def test_when_only_source_is_specified(runner, source, expected):
    source: Path = data_filename_resolver(source)
    with runner.isolated_filesystem():
        result = runner.invoke(mckit, args=["decompose", source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success without specified output: " + result.output
        output: Path = get_default_output_directory(source)
        for f in expected:
            p = output / f
            assert p.exists(), f"Should store the file {p} in the default directory '{output}'"
            model = from_file(p).universe
            for cell in model:
                assert "U" not in cell.options or cell.options["U"].name() == 0


@pytest.mark.parametrize(
    "source,output,expected",
    [("cli/data/simple_cubes.mcnp", "split-1", "envelopes.i u1.i u2.i".split())],
)
def test_when_output_is_specified(runner, source, output, expected):
    source = data_filename_resolver(source)
    with runner.isolated_filesystem():
        output = Path(output)
        assert not output.exists(), f"The {output} directory should not exist before the test run"
        result = runner.invoke(
            mckit,
            args=["decompose", "--output", str(output), source],
            catch_exceptions=False,
        )
        assert result.exit_code == 0
        assert output.exists(), f"The {output} directory should exist after the test run"
        for f in expected:
            assert (
                output / f
            ).exists(), f"Should store the file {f} in the default directory 'universes'"


def test_when_output_file_exists_and_override_is_not_specified(runner):
    source = data_filename_resolver("cli/data/simple_cubes.mcnp")
    with runner.isolated_filesystem() as prefix:
        output = Path(prefix) / "simple_cubes.universes/envelopes.i"
        output.parent.mkdir(parents=True)
        output.touch(exist_ok=False)
        result = runner.invoke(mckit, args=["decompose", source], catch_exceptions=False)
        assert (
            result.exit_code != 0
        ), "Should fail when output file exists and --override is not specified"


def test_when_output_file_exists_and_override_is_specified(runner):
    source = data_filename_resolver("cli/data/simple_cubes.mcnp")
    with runner.isolated_filesystem() as prefix:
        output = Path(prefix) / "simple_cubes./envelopes.i"
        output.parent.mkdir(parents=True)
        output.touch(exist_ok=False)
        result = runner.invoke(
            mckit, args=["--override", "decompose", source], catch_exceptions=False
        )
        assert (
            result.exit_code == 0
        ), "Should success when output file exists and --override is specified"


def test_fill_descriptor(runner):
    source = data_filename_resolver("cli/data/simple_cubes.mcnp")
    with runner.isolated_filesystem() as prefix:
        output = Path(prefix) / "simple_cubes.universes/fill-descriptor.toml"
        result = runner.invoke(mckit, args=["decompose", source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success"
        assert output.exists()
        with output.open() as fid:
            fill_descriptor = fid.read()
            assert fill_descriptor.find("simple_cubes.mcnp")
            fill_descriptor = tomllib.loads(fill_descriptor)
            assert "created" in fill_descriptor
            assert "2" in fill_descriptor
            assert "universe" in fill_descriptor["2"]
            assert 1 == fill_descriptor["2"]["universe"]
            assert "u1.i" == fill_descriptor["2"]["file"]


def test_fill_descriptor_when_fill_descriptor_file_is_specified(runner):
    source = data_filename_resolver("cli/data/simple_cubes.mcnp")
    with runner.isolated_filesystem() as prefix:
        fill_descriptor_path = Path(prefix) / "fill-descriptor-special.toml"
        result = runner.invoke(
            mckit,
            args=["decompose", "--fill-descriptor", str(fill_descriptor_path), source],
            catch_exceptions=False,
        )
        assert result.exit_code == 0, "Should success"
        assert fill_descriptor_path.exists()


def test_anonymous_transformation(runner):
    source = data_filename_resolver("cli/data/cubes_with_fill_transforms.mcnp")
    with runner.isolated_filesystem() as prefix:
        output = Path(prefix) / "cubes_with_fill_transforms.universes"
        result = runner.invoke(mckit, args=["decompose", source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success"
        with open(output / "fill-descriptor.toml", "rb") as fid:
            descriptor = tomllib.load(fid)
            spec = descriptor["2"]["transform"]
            assert len(spec) == 3
            spec1 = np.fromiter(map(float, spec), dtype=float)
            assert_array_equal(spec1, [0.0, -1.0, 0.0], f"Fill descriptor {spec1} is wrong")


def test_named_transformation(runner):
    source = data_filename_resolver("cli/data/cubes_with_fill_named_transforms.mcnp")
    with runner.isolated_filesystem() as prefix:
        output = Path(prefix) / "cubes_with_fill_named_transforms.universes"
        result = runner.invoke(mckit, args=["decompose", source], catch_exceptions=False)
        assert result.exit_code == 0, "Should success"
        with open(output / "fill-descriptor.toml", "rb") as fid:
            descriptor = tomllib.load(fid)
            spec = descriptor["2"]["transform"]
            assert spec == 1, f"Fill descriptor {spec} is wrong"
            transforms = descriptor["named_transformations"]
            assert "tr1" in transforms, "Should store transformation tr1"
            transform = transforms["tr1"]
            transform_params = np.fromiter(map(float, transform), dtype=float)
            assert transform_params.size == 3, "Only translation is specified for tr1"
            assert_array_equal(
                transform_params, [0, -1.0, 0]
            ), f"Invalid transform {transform_params}"
            assert transforms is not None



# -*- coding: utf-8 -*-
import re

import mckit.parser.common.utils as m
import pytest


@pytest.mark.parametrize(
    "text,expected",
    [
        ("a\nc\nb", "a\nb"),
        (
            """m1
 c some comment
        1001.21c -1.0
        """,
            """m1
        1001.21c -1.0
        """,
        ),
        (
            """
(
   a
)
c
(
   b
)
""",
            """
(
   a
)
(
   b
)
""",
        ),
    ],
)
def test_drop_c_comments(text, expected):
    actual = m.drop_c_comments(text)
    assert actual == expected


def test_when_no_c_comments_in_text():
    text = """m1
    1001.21c -1.0
    """
    actual = m.drop_c_comments(text)
    assert actual is text, "drop_c_comments should return the text object without changes"


RE_FLOAT = re.compile(m.FLOAT)


@pytest.mark.parametrize(
    "text,expected",
    [
        ("1", "1"),
        (" 0.1", "0.1"),
        (".2\n", ".2"),
        ("\n -1e10a", "-1e10"),
        ("0.", "0."),
    ],
)
def test_float_pattern(text, expected):
    match = RE_FLOAT.search(text)
    assert match, "Should find float number in the text"
    actual = match.group()
    assert actual == expected


@pytest.mark.parametrize(
    "text, expected_new_text, expected_comments, expected_trailing_comment",
    [
        ("1 $ zzz", "1 ", {1: ("zzz",)}, None),
        ("1 $ zzz\n $ ttt", "1 ", {1: ("zzz",)}, ["ttt"]),
        ("1\n $ ttt", "1", None, ["ttt"]),
        ("1 $ zzz\n $ ttt\n2", "1 \n2", {1: ("zzz", "ttt")}, None),
        (
            """M1000
1001.21c -1.0
    gas 1
    $ trailing comment1
    $ trailing comment2
""",
            """M1000
1001.21c -1.0
    gas 1
""",
            None,
            ["trailing comment1", "trailing comment2"],
        ),
    ],
)
def test_extract_comments(text, expected_new_text, expected_comments, expected_trailing_comment):
    actual_new_text, actual_comments, actual_trailing_comment = m.extract_comments(text)
    assert actual_new_text == expected_new_text
    assert actual_comments == expected_comments
    assert actual_trailing_comment == expected_trailing_comment



from typing import Iterable

import re

C_COMMENT = r"(^|(?<=\n))\s{0,5}[cC]([ ][^\n]*)?\n?"
RE_C_COMMENT = re.compile(C_COMMENT, re.MULTILINE)
EOL_COMMENT = r"\$.*[^\n]*"
RE_EOL_COMMENT = re.compile(EOL_COMMENT, re.MULTILINE)
LINE = r"(?P<text>\s*[^ $][^$]*)?(?:\s*\$\s*(?P<comment>.*))?"  # text should contain at list one non space character
RE_LINE = re.compile(LINE)
# FLOAT = r"[-+]?\d*\.?\d+(?:e[-+]?\d+)?"
FLOAT = r"[+-]?((\d+\.?\d*)|(\.\d+))(?:[ed][-+]?\d+)?"
INTEGER = r"\d+"
RE_EMPTY_LINE = re.compile(r"\s*")


def ensure_lower(text: str):
    if not text.islower():
        text = text.lower()
    return text


def ensure_upper(text: str):
    if not text.isupper():
        text = text.upper()
    return text


def drop_c_comments(text: str) -> str:
    has_comments = RE_C_COMMENT.search(text)
    if has_comments:
        text = RE_C_COMMENT.sub("", text)
    return text


def drop_eol_comments(text):
    if RE_EOL_COMMENT.search(text) is not None:
        text = RE_EOL_COMMENT.sub("", text)
    return text


def drop_comments(text):
    return drop_eol_comments(drop_c_comments(text))


def extract_comments(text):
    lines = text.split("\n")
    cleaned_text = []
    comments = []
    trailing_comment = []

    def add_trailing_to_previous_item():
        nonlocal comments, trailing_comment
        if trailing_comment:
            assert 0 < len(
                comments
            ), "The comments should not be empty on this call: at least card name is read"
            comments[-1][1].extend(trailing_comment)
            trailing_comment = []

    for i, line in enumerate(lines):
        if RE_EMPTY_LINE.fullmatch(line):
            cleaned_text.append(line)
        else:
            match = RE_LINE.fullmatch(line)
            assert match is not None
            groupdict = match.groupdict()
            t = groupdict["text"]
            c = groupdict["comment"]
            if t:
                cleaned_text.append(t)
                add_trailing_to_previous_item()
                if c is not None:
                    comments.append((i + 1, [c]))  # lexer counts lines from 1
            else:
                assert c is not None, "If there's no text, then at least comment should present"
                trailing_comment.append(c)

    assert cleaned_text, "There should be some  text in a card"

    if comments:
        comments = dict((k, tuple(v)) for k, v in comments)
    else:
        comments = None

    if not trailing_comment:
        trailing_comment = None

    return "\n".join(cleaned_text), comments, trailing_comment


class ParseError(ValueError):
    """Parsing exception."""


def internalize(word: str, words: Iterable[str]) -> str:
    """Replaces given `word` with the equal word from the list `words` to reuse the object for repeating small words."""
    for w in words:
        if w == word:
            return w, True
    return word, False



from pathlib import Path

import pytest

from mckit.utils.io import check_if_all_paths_exist, check_if_path_exists, make_dirs

TEST_VAR = "TEST_GET_ROOT_DIR_VAR"


def test_mkdirs_and_check_if_all_paths_exist(cd_tmpdir):
    dirs = [*make_dirs(*(Path(f) for f in ["a", "b"]))]
    for p in dirs:
        assert p.exists()
    existing_dirs = [*check_if_all_paths_exist(*dirs)]
    assert existing_dirs == dirs
    with pytest.raises(FileNotFoundError):
        _ = [*check_if_all_paths_exist(Path("not-existing.and.never-should-exist"))]


def test_chk_path(cd_tmpdir):
    check_if_path_exists(Path())
    with pytest.raises(FileNotFoundError):
        check_if_path_exists(Path("not-existing.and.never-should-exist"))
